import os
import errno
import json
from shutil import rmtree

from .group import Group

# set correct json error type for python 2 / 3
try:
    from json.decoder import JSONDecodeError
except ImportError:
    JSONDecodeError = ValueError


class File(Group):
    """ File to access zarr or n5 containers on disc.

    The container corresponds to a directory on the filesystem.
    Groups are subdirectories and datasets are subdirectories
    that contain multi-dimensional data stored in binary format.
    Supports python dict api.

    Args:
        path (str): path on filesystem that holds the container.
        mode (str): file mode used to open / create the file (default: 'a').
        use_zarr_format (bool): flag to determine if container is zarr or n5 (default: None).
    """

    #: file extensions that are inferred as zarr file
    zarr_exts = {'.zarr', '.zr'}
    #: file extensions that are inferred as n5 file
    n5_exts = {'.n5'}

    @classmethod
    def infer_format(cls, path):
        """ Infer the file format from the file extension.

            Returns:
                bool: `True` for zarr, `False` for n5 and `None` if the format could not be infered.
        """
        # first, try to infer the format from the file ending
        is_zarr = None
        _, ext = os.path.splitext(path)
        if ext.lower() in cls.zarr_exts:
            is_zarr = True
        elif ext.lower() in cls.n5_exts:
            is_zarr = False
        # otherwise, infer from the existence of zarr attribute files
        if is_zarr is None:
            zarr_group = os.path.join(path, '.zgroup')
            zarr_array = os.path.join(path, '.zarray')
            is_zarr = os.path.exists(zarr_group) or os.path.exists(zarr_array)
        return is_zarr

    def __init__(self, path, mode='a', use_zarr_format=None):

        # infer the file format from the path
        is_zarr = self.infer_format(path)
        # check if the format that was infered is consistent with `use_zarr_format`
        if use_zarr_format is None:
            if is_zarr is None:
                raise RuntimeError("Cannot infer the file format (zarr or N5)")
            use_zarr_format = is_zarr

        elif use_zarr_format:
            if not is_zarr:
                raise RuntimeError("N5 file cannot be opened in zarr format")

        else:
            if is_zarr:
                raise RuntimeError("Zarr file cannot be opened in N5 format")
        super(File, self).__init__(path, use_zarr_format, mode)

        # check if the file already exists and load if it does
        if os.path.exists(path):

            # throw error if the file must not exist according to file mode
            if self._permissions.must_not_exist():
                raise OSError(errno.EEXIST, os.strerror(errno.EEXIST), path)

            if self._permissions.should_truncate():
                rmtree(path)
                os.mkdir(path)

            # TODO check zarr version as well
            if not use_zarr_format:
                self._check_n5_version(path)

        # otherwise create a new file
        else:
            if not self._permissions.can_create():
                raise OSError(errno.EROFS, os.strerror(errno.EROFS), path)
            os.mkdir(path)
            meta_file = os.path.join(path, '.zgroup' if use_zarr_format else 'attributes.json')
            # we only need to write meta data for the zarr format
            if use_zarr_format:
                with open(meta_file, 'w') as f:
                    json.dump({'zarr_format': 2}, f)
            else:
                with open(meta_file, 'w') as f:
                    json.dump({'n5': "2.0.0"}, f)

    @staticmethod
    def _check_n5_version(path):
        have_version_tag = False
        attrs_file = os.path.join(path, 'attributes.json')
        # check if we have an attributes
        if os.path.exists(attrs_file):
            with open(attrs_file, 'r') as f:
                try:
                    attrs = json.load(f)
                except JSONDecodeError:
                    attrs = {}
            # check if attributes have a version tag
            if 'n5' in attrs:
                tag = attrs['n5']
                have_version_tag = True
        # TODO check for proper version command with regex
        if have_version_tag:
            major_version = int(tag[0])
            if major_version > 2:
                raise RuntimeError("Can't open n5 file with major version bigger than 2")

    def close(self):
        # This function exists just for conformity with the standard file-handling procedure.
        pass

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass


class N5File(File):
    """ File to access n5 containers on disc.

    Args:
        path (str): path on filesystem that holds the container.
        mode (str): file mode used to open / create the file (default: 'a').
    """

    def __init__(self, path, mode='a'):
        super(N5File, self).__init__(path=path, use_zarr_format=False, mode=mode)


class ZarrFile(File):
    """ File to access zarr containers on disc.

    Args:
        path (str): path on filesystem that holds the container.
        mode (str): file mode used to open / create the file (default: 'a').
    """

    def __init__(self, path, mode='a'):
        super(ZarrFile, self).__init__(path=path, use_zarr_format=True, mode=mode)



import sys
import unittest
import numpy as np
import os
from shutil import rmtree

try:
    from concurrent import futures
except ImportError:
    futures = False

try:
    import h5py
except ImportError:
    h5py = False

try:
    import z5py
except ImportError:
    sys.path.append('..')
    import z5py


class TestConverter(unittest.TestCase):
    tmp_dir = './tmp_dir'
    shape = (100, 100, 100)
    chunks = (10, 10, 10)

    compressor_pairs = [('raw', 'raw'), ('gzip', 'gzip'),
                        ('raw', 'gzip'), ('gzip', 'raw')]

    def setUp(self):
        if not os.path.exists(self.tmp_dir):
            os.mkdir(self.tmp_dir)

    def tearDown(self):
        try:
            rmtree(self.tmp_dir)
        except OSError:
            pass

    @unittest.skipUnless(h5py, 'Requires h5py')
    @unittest.skipUnless(futures, 'Needs 3rd party concurrent.futures in python 2')
    def test_h5_to_n5(self):
        from z5py.converter import convert_from_h5
        h5_file = os.path.join(self.tmp_dir, 'tmp.h5')
        n5_file = os.path.join(self.tmp_dir, 'tmp.n5')

        # test conversion for 4 different compression pairs
        for pair in self.compressor_pairs:

            key = 'data_%s_%s' % pair
            compression_a, compression_b = pair
            compression_a = None if compression_a == 'raw' else compression_a

            with h5py.File(h5_file, 'w') as f:
                ds = f.create_dataset(key,
                                      shape=self.shape,
                                      chunks=self.chunks,
                                      compression=compression_a)
                data = np.arange(ds.size).reshape(ds.shape).astype(ds.dtype)
                ds[:] = data

            convert_from_h5(h5_file, n5_file,
                            key, key, chunks=self.chunks,
                            n_threads=8, compression=compression_b)

            fn5 = z5py.File(n5_file)
            data_n5 = fn5[key][:]
            self.assertTrue(np.allclose(data, data_n5))

    @unittest.skipUnless(h5py, 'Requires h5py')
    @unittest.skipUnless(futures, 'Needs 3rd party concurrent.futures in python 2')
    def test_h5_to_n5_with_roi(self):
        from z5py.converter import convert_from_h5
        h5_file = os.path.join(self.tmp_dir, 'tmp.h5')
        n5_file = os.path.join(self.tmp_dir, 'tmp.n5')

        h5_key = 'data'
        with h5py.File(h5_file, 'w') as f:
            ds = f.create_dataset(h5_key, shape=self.shape, chunks=self.chunks)
            data = np.arange(ds.size).reshape(ds.shape).astype(ds.dtype)
            ds[:] = data

        # define roi
        roi = np.s_[5:45, 9:83, 10:60]
        out_of_roi_mask = np.ones(self.shape, dtype='bool')
        out_of_roi_mask[roi] = False
        roi_shape = tuple(rr.stop - rr.start for rr in roi)

        # convert dataset with roi
        n5_key = 'data_roi'
        convert_from_h5(h5_file, n5_file,
                        h5_key, n5_key, n_threads=8,
                        roi=roi, fit_to_roi=False)

        fn5 = z5py.File(n5_file)
        data_n5 = fn5[n5_key][:]
        self.assertEqual(data.shape, data_n5.shape)
        self.assertTrue(np.allclose(data[roi], data_n5[roi]))
        self.assertTrue(np.allclose(data_n5[out_of_roi_mask], 0))

        # convert dataset with roi and fit_to_roi
        n5_key = 'data_roi_fit'
        convert_from_h5(h5_file, n5_file,
                        h5_key, n5_key, n_threads=8,
                        roi=roi, fit_to_roi=True)

        fn5 = z5py.File(n5_file)
        data_n5 = fn5[n5_key][:]
        self.assertEqual(roi_shape, data_n5.shape)
        self.assertTrue(np.allclose(data[roi], data_n5))

    @unittest.skipUnless(h5py, 'Requires h5py')
    @unittest.skipUnless(futures, 'Needs 3rd party concurrent.futures in python 2')
    def test_n5_to_h5(self):
        from z5py.converter import convert_to_h5
        n5_file = os.path.join(self.tmp_dir, 'tmp.n5')
        h5_file = os.path.join(self.tmp_dir, 'tmp.h5')
        f = z5py.File(n5_file, use_zarr_format=False)

        # test conversion for 4 different compression pairs
        for pair in self.compressor_pairs:

            key = 'data_%s_%s' % pair
            compression_a, compression_b = pair

            ds = f.create_dataset(key,
                                  dtype='float32',
                                  compression=compression_a,
                                  shape=self.shape,
                                  chunks=self.chunks)
            data = np.arange(ds.size).reshape(ds.shape).astype(ds.dtype)
            ds[:] = data

            compression_b = None if compression_b == 'raw' else compression_b
            convert_to_h5(n5_file, h5_file,
                          key, key, chunks=self.chunks,
                          n_threads=1, compression=compression_b)

            with h5py.File(h5_file, 'r') as fh5:
                data_h5 = fh5[key][:]
            self.assertTrue(np.allclose(data, data_h5))

    @unittest.skipUnless(h5py, 'Requires h5py')
    @unittest.skipUnless(futures, 'Needs 3rd party concurrent.futures in python 2')
    def test_n5_to_h5_with_roi(self):
        from z5py.converter import convert_to_h5
        h5_file = os.path.join(self.tmp_dir, 'tmp.h5')
        n5_file = os.path.join(self.tmp_dir, 'tmp.n5')

        n5_key = 'data'
        with z5py.File(n5_file, 'w') as f:
            ds = f.create_dataset(n5_key, shape=self.shape, chunks=self.chunks,
                                  dtype='float64')
            data = np.arange(ds.size).reshape(ds.shape).astype(ds.dtype)
            ds[:] = data

        # define roi
        roi = np.s_[5:45, 9:83, 10:60]
        out_of_roi_mask = np.ones(self.shape, dtype='bool')
        out_of_roi_mask[roi] = False
        roi_shape = tuple(rr.stop - rr.start for rr in roi)

        # convert dataset with roi
        h5_key = 'data_roi'
        convert_to_h5(n5_file, h5_file,
                      n5_key, h5_key, n_threads=8,
                      roi=roi, fit_to_roi=False)

        with h5py.File(h5_file, 'r') as f:
            data_h5 = f[h5_key][:]
        self.assertEqual(data.shape, data_h5.shape)
        self.assertTrue(np.allclose(data[roi], data_h5[roi]))
        self.assertTrue(np.allclose(data_h5[out_of_roi_mask], 0))

        # convert dataset with roi and fit_to_roi
        h5_key = 'data_roi_fit'
        convert_to_h5(n5_file, h5_file,
                      n5_key, h5_key, n_threads=8,
                      roi=roi, fit_to_roi=True)

        with h5py.File(h5_file, 'r') as f:
            data_h5 = f[h5_key][:]
        self.assertEqual(roi_shape, data_h5.shape)
        self.assertTrue(np.allclose(data[roi], data_h5))


if __name__ == '__main__':
    unittest.main()



import unittest
import sys
import numpy as np
import os
from shutil import rmtree

try:
    import zarr
    import numcodecs
    HAVE_ZARR = True
except ImportError:
    HAVE_ZARR = False

try:
    import z5py
except ImportError:
    sys.path.append('..')
    import z5py


class TestZarrCompatibility(unittest.TestCase):
    def setUp(self):
        self.path = 'f.zr'
        self.shape = (100, 100)
        self.chunks = (20, 20)

    def tearDown(self):
        try:
            rmtree(self.path)
        except OSError:
            pass

    @unittest.skipUnless(HAVE_ZARR, 'Requires zarr package')
    def test_read_zarr_irregular(self):
        shape = (123, 97)
        chunks = (17, 32)
        data = np.random.rand(*shape)
        fz = zarr.open(self.path)
        fz.create_dataset('test', data=data, chunks=chunks)

        f = z5py.File(self.path)
        out = f['test'][:]

        self.assertEqual(data.shape, out.shape)
        self.assertTrue(np.allclose(data, out))

    @unittest.skipUnless(HAVE_ZARR, 'Requires zarr package')
    def test_write_zarr_irregular(self):
        shape = (123, 97)
        chunks = (17, 32)
        data = np.random.rand(*shape)
        f = z5py.File(self.path)
        f.create_dataset('test', data=data, chunks=chunks)

        fz = z5py.File(self.path)
        out = fz['test'][:]

        self.assertEqual(data.shape, out.shape)
        self.assertTrue(np.allclose(data, out))

    @unittest.skipUnless(HAVE_ZARR, 'Requires zarr package')
    def test_read_zarr(self):
        from z5py.dataset import Dataset
        dtypes = list(Dataset._zarr_dtype_dict.values())
        zarr_compressors = {'blosc': numcodecs.Blosc(),
                            'zlib': numcodecs.Zlib(),
                            'raw': None,
                            'bzip2': numcodecs.BZ2()}
        # conda-forge version of numcodecs is not up-to-data
        # for python 3.5 and GZip is missing
        # thats why we need to check explicitly here to not fail the test
        if hasattr(numcodecs, 'GZip'):
            zarr_compressors.update({'gzip': numcodecs.GZip()})

        for dtype in dtypes:
            for compression in zarr_compressors:
                data = np.random.randint(0, 127, size=self.shape).astype(dtype)
                # write the data with zarr
                key = 'test_%s_%s' % (dtype, compression)
                ar = zarr.open(os.path.join(self.path, key), mode='w',
                               shape=self.shape, chunks=self.chunks,
                               dtype=dtype, compressor=zarr_compressors[compression])
                ar[:] = data
                # read with z5py
                out = z5py.File(self.path)[key][:]
                self.assertEqual(data.shape, out.shape)
                self.assertTrue(np.allclose(data, out))

    @unittest.skipUnless(HAVE_ZARR, 'Requires zarr package')
    def test_write_zarr(self):
        from z5py.dataset import Dataset
        dtypes = list(Dataset._zarr_dtype_dict.values())
        compressions = Dataset.compressors_zarr

        for dtype in dtypes:
            for compression in compressions:
                data = np.random.randint(0, 127, size=self.shape).astype(dtype)
                key = 'test_%s_%s' % (dtype, compression)
                # read with z5py
                ds = z5py.File(self.path)
                ds.create_dataset(key, data=data, compression=compression,
                                  chunks=self.chunks)
                # read the data with z5py
                ar = zarr.open(os.path.join(self.path, key))
                out = ar[:]
                self.assertEqual(data.shape, out.shape)
                self.assertTrue(np.allclose(data, out))

    @unittest.skipUnless(HAVE_ZARR, 'Requires zarr package')
    def test_fillvalue(self):
        test_values = [0, 10, 42, 255]
        for val in test_values:
            key = 'test_%i' % val
            zarr.open(os.path.join(self.path, key), shape=self.shape,
                      fill_value=val, dtype='<u1')
            out = z5py.File(self.path)[key][:]
            self.assertEqual(self.shape, out.shape)
            self.assertTrue(np.allclose(val, out))


if __name__ == '__main__':
    unittest.main()



# helper functions to convert to and from different formats
# - hdf5, tiff
from __future__ import print_function
import os
from math import ceil
from concurrent import futures
import numpy as np

try:
    import h5py
    WITH_H5 = True
except ImportError:
    print("h5py is not installed, hdf5 converters not available")
    WITH_H5 = False

try:
    import imageio
    WITH_IMIO = True
except ImportError:
    print("imageio is not installed, tif converters not available")
    WITH_IMIO = False

from .file import File
from .util import blocking
from .shape_utils import normalize_slices

if WITH_H5:

    def convert_to_h5(in_path, out_path,
                      in_path_in_file, out_path_in_file,
                      n_threads, chunks=None,
                      block_shape=None, roi=None,
                      fit_to_roi=False, **h5_kwargs):
        """ Convert n5 ot zarr dataset to hdf5 dataset.

        The chunks of the output dataset must be spcified.
        The dataset is converted to hdf5 in parallel over the chunks.
        Note that hdf5 does not support parallel write access, so more threads
        may not speed up the conversion.
        Datatype and compression can be specified, otherwise defaults will be used.

        Args:
            in_path (str): path to n5 or zarr file.
            out_path (str): path to output hdf5 file.
            in_path_in_file (str): name of input dataset.
            out_path_in_file (str): name of output dataset.
            n_threads (int): number of threads used for converting.
            chunks (tuple): chunks of output dataset.
                By default input datase's chunks are used. (default: None)
            block_shape (tuple): block shape used for converting, must be multiple of ``chunks``.
                If None, the chunk size will be used (default: None).
            roi (tuple[slice]): region of interest that will be copied. (default: None)
            fit_to_roi (bool): if given a roi, whether to set the shape of
                the output dataset to the roi's shape
                and align chunks with the roi's origin. (default: False)
            **h5_kwargs: keyword arguments for ``h5py`` dataset, e.g. datatype or compression.
        """
        if not os.path.exists(in_path):
            raise RuntimeError("Path %s does not exist" % in_path)
        f_z5 = File(in_path)
        ds_z5 = f_z5[in_path_in_file]
        shape = ds_z5.shape
        chunks = ds_z5.chunks if chunks is None else chunks
        # modify h5 arguments
        out_dtype = h5_kwargs.pop('dtype', ds_z5.dtype)
        if block_shape is None:
            block_shape = chunks
        else:
            assert all(bs % ch == 0 for bs, ch in zip(block_shape, chunks)),\
                "block_shape must be a multiple of chunks"

        # we need to create the blocking here, before the shape is potentially altered
        # if fit_to_roi == True
        blocks = blocking(shape, block_shape, roi, fit_to_roi)
        if roi is not None:
            roi, _ = normalize_slices(roi, shape)
            if fit_to_roi:
                shape = tuple(rr.stop - rr.start for rr in roi)

        with h5py.File(out_path) as f_h5:
            ds_h5 = f_h5.require_dataset(out_path_in_file,
                                         dtype=out_dtype,
                                         shape=shape,
                                         chunks=chunks,
                                         **h5_kwargs)

            def convert_chunk(bb):
                # print("Converting chunk ", chunk_ids, "/", chunks_per_dim)
                # don't copy empty data
                chunk_data = ds_z5[bb].astype(out_dtype, copy=False)
                if chunk_data.sum() == 0:
                    return
                # if we have a roi and fit to it, we need to substract its start from
                # the current bounding box
                if fit_to_roi and roi is not None:
                    bb = tuple(slice(b.start - rr.start, b.stop - rr.start)
                               for b, rr in zip(bb, roi))
                ds_h5[bb] = chunk_data

            with futures.ThreadPoolExecutor(max_workers=n_threads) as tp:
                tasks = [tp.submit(convert_chunk, bb) for bb in blocks]
                [t.result() for t in tasks]

            # copy attributes
            h5_attrs = ds_h5.attrs
            z5_attrs = ds_z5.attrs
            for key, val in z5_attrs.items():
                h5_attrs[key] = val


    def convert_from_h5(in_path, out_path,
                        in_path_in_file, out_path_in_file,
                        n_threads, chunks=None,
                        block_shape=None, use_zarr_format=None,
                        roi=None, fit_to_roi=False,
                        **z5_kwargs):
        """ Convert hdf5 dataset to n5 or zarr dataset.

        The chunks of the output dataset must be spcified.
        The dataset is converted in parallel over the chunks.
        Datatype and compression can be specified, otherwise defaults will be used.

        Args:
            in_path (str): path to hdf5 file.
            out_path (str): path to output zarr or n5 file.
            in_path_in_file (str): name of input dataset.
            out_path_in_file (str): name of output dataset.
            n_threads (int): number of threads used for converting.
            chunks (tuple): chunks of output dataset.
             By default input dataset's chunks are used. (default: None)
            block_shape (tuple): block shape used for converting, must be multiple of ``chunks``.
                If None, the chunk shape will be used (default: None).
            use_zarr_format (bool): flag to indicate zarr format.
                If None, an attempt will be made to infer the format from the file extension,
                otherwise zarr will be used (default: None).
            roi (tuple[slice]): region of interest that will be copied. (default: None)
            fit_to_roi (bool): if given a roi, whether to set the shape of
                the output dataset to the roi's shape
                and align chunks with the roi's origin. (default: False)
            **z5_kwargs: keyword arguments for ``z5py`` dataset, e.g. datatype or compression.
        """
        if not os.path.exists(in_path):
            raise RuntimeError("Path %s does not exist" % in_path)

        f_z5 = File(out_path, use_zarr_format=use_zarr_format)
        with h5py.File(in_path, 'r') as f_h5:
            ds_h5 = f_h5[in_path_in_file]
            shape = ds_h5.shape
            chunks = ds_h5.chunks if chunks is None else chunks

            if block_shape is None:
                block_shape = chunks
            else:
                assert all(bs % ch == 0 for bs, ch in zip(block_shape, chunks)),\
                    "block_shape must be a multiple of chunks"

            # we need to create the blocking here, before the shape is potentially altered
            # if fit_to_roi == True
            blocks = blocking(shape, block_shape, roi, fit_to_roi)
            if roi is not None:
                roi, _ = normalize_slices(roi, shape)
                if fit_to_roi:
                    shape = tuple(rr.stop - rr.start for rr in roi)

            # modify z5 arguments
            out_dtype = z5_kwargs.pop('dtype', ds_h5.dtype)
            if 'compression' not in z5_kwargs:
                z5_kwargs['compression'] = 'raw'
            ds_z5 = f_z5.require_dataset(out_path_in_file,
                                         dtype=out_dtype,
                                         shape=shape,
                                         chunks=chunks,
                                         **z5_kwargs)

            def convert_chunk(bb):
                chunk_data = ds_h5[bb].astype(out_dtype, copy=False)
                # if we have a roi and fit to it, we need to substract its start from
                # the current bounding box
                if fit_to_roi and roi is not None:
                    bb = tuple(slice(b.start - rr.start, b.stop - rr.start)
                               for b, rr in zip(bb, roi))
                ds_z5[bb] = chunk_data

            with futures.ThreadPoolExecutor(max_workers=n_threads) as tp:
                tasks = [tp.submit(convert_chunk, bb) for bb in blocks]
                [t.result() for t in tasks]

            # copy attributes
            h5_attrs = ds_h5.attrs
            z5_attrs = ds_z5.attrs
            for key, val in h5_attrs.items():
                # h5 attributes might come as ndarrays, which are not
                # json deserializable by default.
                # in this case, cast to list
                if isinstance(val, np.ndarray):
                    val = val.tolist()
                z5_attrs[key] = val


if WITH_IMIO:

    def convert_to_tif():
        raise NotImplementedError("Conversion to tif not implemented.")


    def is_int(string):
        try:
            int(string)
            return True
        except ValueError:
            return False


    # TODO use proper regex
    def default_index_parser(fname):
        # get rid of the postfix
        parsed = os.path.splitext(fname)[0]
        # try to convert to interger directly
        if is_int(parsed):
            return int(parsed)
        # try to split at '_'
        parsed = parsed.split('_')
        if len(parsed) == 1:
            # otherwise try to split at '-'
            parsed = parsed[0].split('-')
            if len(parsed) == 1 and not is_int(parsed[0]):
                raise ValueError("Cannot parse file with default parser")
        # convert the parsed strings to int
        parsed = [int(par) for par in parsed if is_int(par)]
        if len(parsed) == 0:
            raise ValueError("Cannot parse file with default parser")
        else:
            # we assume that the image-index is always the last number parsed
            return parsed[-1]


    def _read_tif_metadata(in_path, file_names=None):
        if file_names is None:
            pass  # TODO
        else:
            # TODO can we somehow read shape from the metadata ???
            # with imageio.get_reader(os.path.join(in_path, file_names[0])) as f:
            #     meta = f.get_meta_data()
            #     print(meta)
            d = imageio.imread(os.path.join(in_path, file_names[0]))
            # TODO check shape for all images once we can read from metadata
            shape, dtype = d.shape, d.dtype
            shape = (len(file_names),) + shape
        return shape, dtype


    # TODO
    # - implement for tif volumes / stacks / multi-page tifs
    # - fix behaviour for multi-channel images: now we would get (z, c, x, y)
    #   but we want (c, z, y x)
    def convert_from_tif(in_path, out_path,
                         out_path_in_file, chunks,
                         n_threads, use_zarr_format=None,
                         parser=None, preprocess=None,
                         **z5_kwargs):
        """ Convert tif stack or folder of tifs to n5 or zarr dataset.

        The chunks of the output dataset must be specified.
        The dataset is converted in parallel over the chunks.
        Datatype and compression can be specified, otherwise defaults will be used.

        Args:
            in_path (str): path to tif stack or folder of tifs.
            out_path (str): path to output zarr or n5 file.
            out_path_in_file (str): name of output dataset.
            chunks (tuple): chunks of output dataset.
            n_threads (int): number of threads used for converting.
            use_zarr_format (bool): flag to indicate zarr format.
                If None, an attempt will be made to infer the format from the file extension,
                otherwise zarr will be used (default: None).
            parser (callable): function to parse the image indices for tifs in a folder.
                If None, some default patterns are tried (default: None)
            process (callable): function to preprocess chunks before wrting to n5/zarr
                Must take np.ndarray and int as arguments. (default: None)
            **z5_kwargs: keyword arguments for ``z5py`` dataset, e.g. datatype or compression.
        """
        parser_ = default_index_parser if parser is None else parser
        if preprocess is not None:
            assert callable(preprocess)

        if os.path.isdir(in_path):
            file_names = [pp for pp in os.listdir(in_path)
                          if pp.split('.')[-1].lower() in ('tif', 'tiff')]
            indices = [parser_(pp) for pp in file_names]
            assert len(indices) == len(file_names)
            # TODO should we check that indices are consecutive and unique ?
            # order file names according to indices
            file_names = [fname for _, fname in sorted(zip(indices, file_names), key=lambda pair: pair[0])]
        elif os.path.isfile(in_path):
            file_names = None
            raise NotImplementedError("Single tiff conversion not implemented")
        else:
            raise RuntimeError("Path %s does not exist" % in_path)

        # get shape and dtype from metadata
        shape, dtype_ = _read_tif_metadata(in_path, file_names)
        print(shape, dtype_)

        # create the z5 file
        f_z5 = File(out_path, use_zarr_format=use_zarr_format)
        dtype = z5_kwargs.pop('dtype', dtype_)

        # create the dataset
        ds_z5 = f_z5.create_dataset(out_path_in_file, dtype=dtype,
                                    shape=shape, chunks=chunks, **z5_kwargs)

        # TODO implement for tif volume
        # write all chunks overlapping with z images
        def convert_block(i0, i1):
            cshape = (i1 - i0,) + shape[1:]
            slice_ = (slice(i0, i1),) + len(shape[1:]) * (slice(None),)
            chunk = np.zeros(cshape, dtype=dtype)
            for z, i in enumerate(range(i0, i1)):
                chunk[z] = imageio.imread(os.path.join(in_path, file_names[i]))\
                    if preprocess is None else preprocess(imageio.imread(os.path.join(in_path, file_names[i])), i)

            ds_z5[slice_] = chunk.astype(dtype, copy=False)

        n_blocks = int(ceil(shape[0] / chunks[0]))
        starts = [chunks[0] * i for i in range(n_blocks)]
        stops = [min(chunks[0] * (i + 1), shape[0]) for i in range(n_blocks)]

        with futures.ThreadPoolExecutor(max_workers=n_threads) as tp:
            tasks = [tp.submit(convert_block, i0, i1)
                     for i0, i1 in zip(starts, stops)]
            [t.result() for t in tasks]



from __future__ import print_function
import os
from itertools import product
from concurrent import futures
from contextlib import closing
from datetime import datetime
import numpy as np

from . import _z5py as z5_impl
from .file import File
from .group import Group
from .dataset import Dataset
from .shape_utils import normalize_slices


def product1d(inrange):
    for ii in inrange:
        yield ii


def blocking(shape, block_shape, roi=None, center_blocks_at_roi=False):
    """ Generator for nd blocking.

    Args:
        shape (tuple): nd shape
        block_shape (tuple): nd block shape
        roi (tuple[slice]): region of interest (default: None)
        center_blocks_at_roi (bool): if given a roi,
            whether to center the blocks being generated
            at the roi's origin (default: False)
    """
    assert len(shape) == len(block_shape), "Invalid number of dimensions."

    if roi is None:
        # compute the ranges for the full shape
        ranges = [range(sha // bsha if sha % bsha == 0 else sha // bsha + 1)
                  for sha, bsha in zip(shape, block_shape)]
        min_coords = [0] * len(shape)
        max_coords = shape
    else:
        # make sure that the roi is valid
        roi, _ = normalize_slices(roi, shape)
        ranges = [range(rr.start // bsha,
                        rr.stop // bsha if rr.stop % bsha == 0 else rr.stop // bsha + 1)
                  for rr, bsha in zip(roi, block_shape)]
        min_coords = [rr.start for rr in roi]
        max_coords = [rr.stop for rr in roi]

    need_shift = False
    if roi is not None and center_blocks_at_roi:
        shift = [rr.start % bsha for rr, bsha in zip(roi, block_shape)]
        need_shift = sum(shift) > 0

    # product raises memory error for too large ranges,
    # because input iterators are cast to tuple
    # TODO so far I have only seen this for 1d "open-ended" datasets
    # and hence just implemented a workaround for this case,
    # but it should be fairly easy to implement an nd version of product
    # without casting to tuple for our use case using the imglib loop trick, see also
    # https://stackoverflow.com/questions/8695422/why-do-i-get-a-memoryerror-with-itertools-product
    try:
        start_points = product(*ranges)
    except MemoryError:
        assert len(ranges) == 1
        start_points = product1d(ranges)

    for start_point in start_points:
        positions = [sp * bshape for sp, bshape in zip(start_point, block_shape)]
        if need_shift:
            positions = [pos + sh for pos, sh in zip(positions, shift)]
            if any(pos > maxc for pos, maxc in zip(positions, max_coords)):
                continue
        yield tuple(slice(max(pos, minc), min(pos + bsha, maxc))
                    for pos, bsha, minc, maxc in zip(positions, block_shape,
                                                     min_coords, max_coords))


def copy_dataset(in_path, out_path,
                 in_path_in_file, out_path_in_file,
                 n_threads, chunks=None,
                 block_shape=None, dtype=None,
                 use_zarr_format=None, roi=None,
                 fit_to_roi=False, **new_compression):
    """ Copy dataset, optionally change metadata.

    The input dataset will be copied to the output dataset chunk by chunk.
    Allows to change chunks, datatype, file format and compression.
    Can also just copy a roi.

    Args:
        in_path (str): path to the input file.
        out_path (str): path to the output file.
        in_path_in_file (str): name of input dataset.
        out_path_in_file (str): name of output dataset.
        n_threads (int): number of threads used for copying.
        chunks (tuple): chunks of the output dataset.
            By default same as input dataset's chunks. (default: None)
        block_shape (tuple): block shape used for copying. Must be a multiple
            of ``chunks``, which are used by default (default: None)
        dtype (str): datatype of the output dataset, default does not change datatype (default: None).
        use_zarr_format (bool): file format of the output file,
            default does not change format (default: None).
        roi (tuple[slice]): region of interest that will be copied. (default: None)
        fit_to_roi (bool): if given a roi, whether to set the shape of
            the output dataset to the roi's shape
            and align chunks with the roi's origin. (default: False)
        **new_compression: compression library and options for output dataset. If not given,
            the same compression as in the input is used.

    """
    f_in = File(in_path)
    # check if the file format was specified
    # if not, keep the format of the input file
    # otherwise set the file format
    is_zarr = f_in.is_zarr if use_zarr_format is None else use_zarr_format
    f_out = File(out_path, use_zarr_format=is_zarr)

    ds_in = f_in[in_path_in_file]

    # check if we can copy chunk by chunk
    if (chunks is None or chunks == ds_in.chunks) and (roi is None):
        copy_chunks = True
    else:
        copy_chunks = False

    # get dataset metadata from input dataset if defaults were given
    chunks = ds_in.chunks if chunks is None else chunks
    dtype = ds_in.dtype if dtype is None else dtype
    compression_opts = new_compression if new_compression else ds_in.compression_opts

    # if we don't have block-shape explitictly given, use chunk size
    # otherwise check that it's a multiple of chunks
    if block_shape is None:
        block_shape = chunks
    else:
        assert all(bs % ch == 0 for bs, ch in zip(block_shape, chunks)),\
            "block_shape must be a multiple of chunks"

    shape = ds_in.shape
    # we need to create the blocking here, before the shape is potentially altered
    # if fit_to_roi == True
    blocks = blocking(shape, block_shape, roi, fit_to_roi)
    if roi is not None:
        roi, _ = normalize_slices(roi, shape)
        if fit_to_roi:
            shape = tuple(rr.stop - rr.start for rr in roi)

    ds_out = f_out.require_dataset(out_path_in_file,
                                   dtype=dtype,
                                   shape=shape,
                                   chunks=chunks,
                                   **compression_opts)

    def write_single_block(bb):
        data_in = ds_in[bb].astype(dtype, copy=False)
        if np.sum(data_in) == 0:
            return
        if fit_to_roi and roi is not None:
            bb = tuple(slice(b.start - rr.start, b.stop - rr.start)
                       for b, rr in zip(bb, roi))
        ds_out[bb] = data_in

    def write_single_chunk(bb):
        chunk_id = tuple(b.start // ch for b, ch in zip(bb, chunks))
        chunk_in = ds_in.read_chunk(chunk_id)
        if chunk_in is None:
            return
        # check if this is a varlen chunk
        varlen = tuple(chunk_in.shape) != tuple(b.stop - b.start for b in bb)
        ds_out.write_chunk(chunk_id, chunk_in.astype(dtype, copy=False), varlen)

    write_single = write_single_chunk if copy_chunks else write_single_block

    with futures.ThreadPoolExecutor(max_workers=n_threads) as tp:
        tasks = [tp.submit(write_single, bb) for bb in blocks]
        [t.result() for t in tasks]

    # copy attributes
    in_attrs = ds_in.attrs
    out_attrs = ds_out.attrs
    for key, val in in_attrs.items():
        out_attrs[key] = val


# TODO we could allow to change formats from zarr to n5 and vice versa
def copy_group(in_path, out_path, in_path_in_file, out_path_in_file, n_threads):
    """ Copy group recursively.

    Copy the group recursively, using copy_dataset. Metadata of datasets that
    are copied cannot be changed and rois cannot be applied.

    Args:
        in_path (str): path to the input file.
        out_path (str): path to the output file.
        in_path_in_file (str): name of input group.
        out_path_in_file (str): name of output group.
        n_threads (int): number of threads used to copy datasets.
    """
    f_in = File(in_path)
    f_out = File(out_path)

    g_in = f_in[in_path_in_file]
    for key, obj in g_in.items():
        abs_in_key = os.path.join(in_path_in_file, key)
        abs_out_key = os.path.join(out_path_in_file, key)
        if isinstance(obj, Dataset):
            copy_dataset(in_path, out_path,
                         abs_in_key, abs_out_key, n_threads)
        else:
            assert isinstance(obj, Group)
            # copy group attributes
            g_out = f_out.require_group(abs_out_key)
            in_attrs = obj.attrs
            out_attrs = g_out.attrs
            for key, val in in_attrs.items():
                out_attrs[key] = val

            # copy group content
            copy_group(in_path, out_path,
                       abs_in_key, abs_out_key, n_threads)


class Timer:
    def __init__(self):
        self.start_time = None
        self.stop_time = None

    @property
    def elapsed(self):
        try:
            return (self.stop_time - self.start_time).total_seconds()
        except TypeError as e:
            if "'NoneType'" in str(e):
                raise RuntimeError("{} either not started, or not stopped".format(self))

    def start(self):
        self.start_time = datetime.utcnow()

    def stop(self):
        self.stop_time = datetime.utcnow()
        return self.elapsed

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop()


def fetch_test_data_stent():
    from imageio import volread
    data_i16 = volread('imageio:stent.npz')
    return (data_i16 / data_i16.max() * 255).astype(np.uint8)


def fetch_test_data():
    try:
        from urllib.request import urlopen
    except ImportError:
        from urllib2 import urlopen

    try:
        from io import BytesIO as Buffer
    except ImportError:
        from StringIO import StringIO as Buffer

    import zipfile
    from imageio import volread

    im_url = "https://imagej.nih.gov/ij/images/t1-head-raw.zip"

    with closing(urlopen(im_url)) as response:
        if response.status != 200:
            raise RuntimeError("Test data could not be found at {}, status code {}".format(
                im_url, response.status
            ))
        zip_buffer = Buffer(response.read())

    with zipfile.ZipFile(zip_buffer) as zf:
        tif_buffer = Buffer(zf.read('JeffT1_le.tif'))
        return np.asarray(volread(tif_buffer, format='tif'), dtype=np.uint8)


def remove_trivial_chunks(dataset, n_threads,
                          remove_specific_value=None):
    """ Remove chunks that only contain a single value.

    The input dataset will be copied to the output dataset chunk by chunk.
    Allows to change datatype, file format and compression as well.

    Args:
        dataset (z5py.Dataset)
        n_threads (int): number of threads
        remove_specific_value (int or float): only remove chunks that contain (only) this specific value (default: None)

    """

    dtype = dataset.dtype
    function = eval('z5_impl.remove_trivial_chunks_%s' % dtype)
    remove_specific = remove_specific_value is not None
    value = remove_specific_value if remove_specific else 0
    function(dataset._impl, n_threads, remove_specific, value)


def remove_dataset(dataset, n_threads):
    """ Remvoe dataset multi-threaded.
    """
    z5_impl.remove_dataset(dataset._impl, n_threads)


def remove_chunk(dataset, chunk_id):
    """ Remove a chunk
    """
    z5_impl.remove_chunk(dataset._impl, chunk_id)


def remove_chunks(dataset, bounding_box):
    """ Remove all chunks overlapping the bounding box
    """
    shape = dataset.shape
    chunks = dataset.chunks
    blocks = blocking(shape, chunks, roi=bounding_box)
    for block in blocks:
        chunk_id = tuple(b.start // ch for b, ch in zip(block, chunks))
        remove_chunk(dataset, chunk_id)


def unique(dataset, n_threads, return_counts=False):
    """ Find unique values in dataset.

    Args:
        dataset (z5py.Dataset)
        n_threads (int): number of threads
        return_counts (bool): return counts of unique values (default: False)

    """
    dtype = dataset.dtype
    if return_counts:
        function = eval('z5_impl.unique_with_counts_%s' % dtype)
    else:
        function = eval('z5_impl.unique_%s' % dtype)
    return function(dataset._impl, n_threads)



from .file import File, N5File, ZarrFile
from .dataset import Dataset
from .group import Group
from .attribute_manager import set_json_encoder, set_json_decoder

__version__ = "1.5.0"
__version_info__ = tuple(int(i) for i in  __version__.split("."))




import json
import os
import errno
from functools import wraps
from contextlib import contextmanager

try:
    from collections.abc import MutableMapping
except ImportError:
    from collections import MutableMapping


__all__ = ['AttributeManager']


# json encoders and decoder
# if None, the default encoder / decoders are used
_JSON_ENCODER = None
_JSON_DECODER = None


def set_json_encoder(encoder):
    """ Set the encoder to be used in `json.dump` for attribute serialization.
    """
    global _JSON_ENCODER
    if encoder is None:
        _JSON_ENCODER = None
        return
    if not isinstance(encoder(), json.JSONEncoder):
        raise RuntimeError("Encoder must inherit from JSONEncoder")
    _JSON_ENCODER = encoder


def get_json_encoder():
    return _JSON_ENCODER


def set_json_decoder(decoder):
    """ Set the decoder to be used in `json.load` for attribute de-serialization.
    """
    global _JSON_DECODER
    if decoder is None:
        _JSON_DECODER = None
        return
    if not isinstance(decoder(), json.JSONDecoder):
        raise RuntimeError("Decoder must inherit from JSONDecoder")
    _JSON_DECODER = decoder


def restrict_metadata_keys(fn):
    """ Decorator for AttributeManager methods which checks that,
    if the manager is for N5, the key argument is not a
    reserved metadata name.
    """
    @wraps(fn)
    def wrapped(obj, key, *args, **kwargs):
        if key in obj.reserved_keys:
            raise RuntimeError("Reserved metadata (key {}) cannot be mutated".format(repr(key)))
        return fn(obj, key, *args, **kwargs)
    return wrapped


class AttributeManager(MutableMapping):
    """ Provides access to custom user attributes.

    Attributes will be saved as json in `attributes.json` for n5, `.zattributes` for zarr.
    Supports the default python dict api.
    N5 stores the dataset attributes in the same file; these attributes are
    NOT mutable via the AttributeManager.
    """
    _zarr_fname = '.zattributes'
    _n5_fname = 'attributes.json'
    _n5_keys = frozenset(('dimensions', 'blockSize',
                          'dataType', 'compressionType', 'compression'))

    def __init__(self, path, is_zarr):
        self.path = os.path.join(path, self._zarr_fname if is_zarr else self._n5_fname)
        self.reserved_keys = frozenset() if is_zarr else self._n5_keys

    def __getitem__(self, key):
        with self._open_attributes() as attributes:
            return attributes[key]

    @restrict_metadata_keys
    def __setitem__(self, key, item):
        with self._open_attributes(True) as attributes:
            attributes[key] = item

    @restrict_metadata_keys
    def __delitem__(self, key):
        with self._open_attributes(True) as attributes:
            del attributes[key]

    @contextmanager
    def _open_attributes(self, write=False):
        """Context manager for JSON attribute store.

        Set ``write`` to True to dump out changes."""
        attributes = self._read_attributes()
        hidden_attrs = {key: attributes.pop(key)
                        for key in self.reserved_keys.intersection(attributes)}
        yield attributes
        if write:
            hidden_attrs.update(attributes)
            self._write_attributes(hidden_attrs)

    def _read_attributes(self):
        """Return dict from JSON attribute store.
        Caller needs to distinguish between valid and N5 metadata keys."""
        global _JSON_DECODER
        try:
            with open(self.path, 'r') as f:
                attributes = json.load(f, cls=_JSON_DECODER)
        except ValueError:
            attributes = {}
        except IOError as e:
            if e.errno == errno.ENOENT:
                attributes = {}
            else:
                raise

        return attributes

    def _write_attributes(self, attributes):
        """Dump ``attributes`` to JSON. Potentially dangerous for N5 metadata."""
        global _JSON_ENCODER
        with open(self.path, 'w') as f:
            json.dump(attributes, f, cls=_JSON_ENCODER)

    def __iter__(self):
        with self._open_attributes() as attributes:
            return iter(attributes)

    def __len__(self):
        with self._open_attributes() as attributes:
            return len(attributes)



import unittest
import sys
import numpy as np
from shutil import rmtree
from abc import ABC

try:
    import z5py
except ImportError:
    sys.path.append('..')
    import z5py


class CompressionTestMixin(ABC):
    def setUp(self):
        self.shape = (100, 100, 100)
        self.root_file = z5py.File('array.' + self.data_format, use_zarr_format=self.data_format == 'zarr')

        self.dtypes = [
            'int8', 'int16', 'int32', 'int64',
            'uint8', 'uint16', 'uint32', 'uint64',
            'float32', 'float64'
        ]

    def tearDown(self):
        try:
            rmtree('array.' + self.data_format)
        except OSError:
            pass

    def check_array(self, result, expected, msg=None):
        self.assertEqual(result.shape, expected.shape, msg)
        self.assertTrue(np.allclose(result, expected), msg)

    def test_compression(self):
        from z5py.dataset import Dataset
        f = self.root_file
        compressions = Dataset.compressors_n5 if self.data_format == 'n5' else\
            Dataset.compressors_zarr

        # iterate over the compression libraries
        for compression in compressions:
            for dtype in self.dtypes:

                ds_name = "ds_%s_%s" % (compression, dtype)
                # generate random data scaled to the range of the current datatype
                min_val, max_val = self.dtype_min_max(dtype)
                in_array = np.random.rand(*self.shape).astype(dtype)
                np.multiply(in_array, max_val, casting='unsafe')
                in_array += min_val

                # FIXME not all compression methods in travis yet
                try:
                    ds = f.create_dataset(ds_name,
                                          data=in_array,
                                          chunks=(10, 10, 10),
                                          compression=compression)
                    out_array = ds[:]
                    self.check_array(out_array, in_array,
                                     'failed for compression %s, dtype %s, format %s' % (compression,
                                                                                         dtype,
                                                                                         self.data_format))
                except RuntimeError:
                    print("Compression", compression, "not found!")
                    continue

    def dtype_min_max(self, dt):
        dtype = np.dtype(dt)
        if issubclass(dtype.type, np.integer):
            info = np.iinfo(dtype)
        else:
            info = np.finfo(dtype)
        return info.min + 1, info.max - 1

    def test_large_values_gzip(self):
        f = self.root_file
        compression = 'zlib' if f.is_zarr else 'gzip'
        size = np.product(self.shape)
        for dtype in self.dtypes:
            ds_name = "ds_%s" % dtype
            min_val, max_val = self.dtype_min_max(dtype)
            # FIXME this causes an overflow warning for float64
            in_array = np.linspace(min_val, max_val, size).astype('uint8')
            in_array = in_array.reshape(self.shape)
            np.random.shuffle(in_array)
            ds = f.create_dataset(ds_name,
                                  data=in_array,
                                  chunks=(10, 10, 10),
                                  compression=compression)
            out_array = ds[:]
            self.check_array(out_array, in_array,
                             'gzip large values failed for, dtype %s, format %s' % (dtype,
                                                                                    self.data_format))

    # this fails for different minimal chunk sizes depending on the
    # zlib version. For now we skip this to not make unittest fail
    # when this version changes.
    # FIXME unittest skip is broken in python 2.7 for mixin pattern
    # @unittest.skip
    def _test_small_chunks_gzip(self):
        f = self.root_file
        # 22 is the chunk-size at which gzip compression fails
        # (interestingly this is NOT the case for zlib encoding)
        shape = (22,)
        chunks = shape
        compression = 'zlib' if f.is_zarr else 'gzip'
        for dtype in self.dtypes:
            ds_name = "ds_%s" % dtype
            ds = f.create_dataset(ds_name, shape=shape, chunks=chunks,
                                  compression=compression, dtype=dtype)
            # generate random data scaled to the range of the current datatype
            min_val, max_val = self.dtype_min_max(dtype)
            in_array = np.random.rand(*shape).astype(dtype)
            np.multiply(in_array, max_val, casting='unsafe')
            in_array += min_val

            ds[:] = in_array
            out = ds[:]
            self.check_array(out, in_array,
                             'gzip small chunks failed for, dtype %s, format %s' % (dtype,
                                                                                    self.data_format))


class TestZarrCompression(CompressionTestMixin, unittest.TestCase):
    data_format = 'zarr'


class TestN5Compression(CompressionTestMixin, unittest.TestCase):
    data_format = 'n5'


if __name__ == '__main__':
    unittest.main()



import sys
import unittest

from shutil import rmtree
from abc import ABC

import numpy as np

try:
    import z5py
except ImportError:
    sys.path.append('..')
    import z5py
import z5py.util


class RegressionTestMixin(ABC):
    @classmethod
    def setUpClass(cls):
        cls.data = z5py.util.fetch_test_data_stent()

    def setUp(self):
        self.root_file = z5py.File('array.' + self.data_format)
        self.chunks = (64,) * 3
        self.n_reps = 10

    def tearDown(self):
        try:
            rmtree('array.' + self.data_format)
        except OSError:
            pass

    def _test_regression_read(self):
        for compression, expected in self.compressions_read.items():
            key = 'ds_%s' % compression
            ds = self.root_file.create_dataset(key, data=self.data, chunks=self.chunks,
                                               compression=compression)
            times = []
            for _ in range(self.n_reps):
                with z5py.util.Timer() as t:
                    ds[:]
                times.append(t.elapsed)
            mint = np.min(times)
            self.assertLess(mint, expected)
            # maxt = np.max(times)
            # print(self.data_format, compression, maxt)

    def _test_regression_write(self):
        for compression, expected in self.compressions_write.items():
            key = 'ds_%s' % compression
            times = []
            for _ in range(self.n_reps):
                with z5py.util.Timer() as t:
                    self.root_file.create_dataset(
                        key, data=self.data, chunks=self.chunks, compression=compression
                    )
                times.append(t.elapsed)
                del self.root_file[key]
            mint = np.min(times)
            self.assertLess(mint, expected)
            # maxt = np.max(times)
            # print(self.data_format, compression, maxt)


class TestN5Regression(RegressionTestMixin, unittest.TestCase):
    data_format = 'n5'
    compressions_read = {'raw': 0.005, 'gzip': 0.04}
    compressions_write = {'raw': 0.006, 'gzip': 0.4}


class TestZarrRegression(RegressionTestMixin, unittest.TestCase):
    data_format = 'zarr'
    compressions_read = {'raw': 0.004, 'zlib': 0.05}
    compressions_write = {'raw': 0.005, 'zlib': 0.4}


if __name__ == '__main__':
    unittest.main()



import os
import numbers
import json


def slice_to_start_stop(s, size):
    """For a single dimension with a given size, normalize slice to size.
     Returns slice(None, 0) if slice is invalid."""
    if s.step not in (None, 1):
        raise ValueError('Nontrivial steps are not supported')

    if s.start is None:
        start = 0
    elif -size <= s.start < 0:
        start = size + s.start
    elif s.start < -size or s.start >= size:
        return slice(None, 0)
    else:
        start = s.start

    if s.stop is None or s.stop > size:
        stop = size
    elif s.stop < 0:
        stop = (size + s.stop)
    else:
        stop = s.stop

    if stop < 1:
        return slice(None, 0)

    return slice(start, stop)


def int_to_start_stop(i, size):
    """For a single dimension with a given size, turn an int into slice(start, stop)
    pair."""
    if -size < i < 0:
        start = i + size
    elif i >= size or i < -size:
        raise ValueError('Index ({}) out of range (0-{})'.format(i, size - 1))
    else:
        start = i
    return slice(start, start + 1)


def sliding_window(arr, wsize):
    """Yield a wsize-length tuple of items from sequence arr as a sliding window"""
    if wsize > len(arr):
        return
    for idx in range(len(arr) - wsize + 1):
        yield tuple(arr[idx:idx + wsize])


def rectify_shape(arr, required_shape):
    """Reshape arr into the required shape while keeping neighbouring non-singleton dimensions together
    e.g. shape (1, 2, 1) -> (2, 1, 1, 1) is fine
    shape (1, 2, 1, 2, 1, 1, 1) -> (1, 2, 2, 1) is not
    """
    if arr.shape == required_shape:
        return arr

    important_shape = list(arr.shape)
    while len(important_shape) > 1:
        if important_shape[0] == 1:
            important_shape.pop(0)
        elif important_shape[-1] == 1:
            important_shape.pop()
        else:
            break

    important_shape = tuple(important_shape)

    msg = ('could not broadcast input array from shape {} into shape {}; '
           'complicated broacasting not supported').format(arr.shape, required_shape)

    if len(important_shape) > len(required_shape):
        raise ValueError(msg)

    is_leading = True
    for window in sliding_window(required_shape, len(important_shape)):
        if window == important_shape:
            if is_leading:
                is_leading = False
                continue
            else:
                break
        if is_leading:
            if window[0] != 1:
                break
        else:
            if window[-1] != 1:
                break
    else:
        return arr.reshape(required_shape)

    raise ValueError(msg)


# TODO we should adjust chunk sizes better if some dimensions
# are larger than the default chunks (e.g. shape (2, 2000, 2000)
# should have chunks (1, 512, 512) instead of (2, 64, 64))
def get_default_chunks(shape):
    # the default size is 64**3
    default_size = 262144
    default_dim = int(round(default_size ** (1. / len(shape))))
    return tuple(min(default_dim, sh) for sh in shape)


def is_group(path, is_zarr):
    if is_zarr:
        return os.path.exists(os.path.join(path, '.zgroup'))
    else:
        meta_path = os.path.join(path, 'attributes.json')
        if not os.path.exists(meta_path):
            return True
        with open(meta_path, 'r') as f:
            # attributes for n5 file can be empty which cannot be parsed by json
            try:
                attributes = json.load(f)
            except ValueError:
                attributes = {}
        # The dimensions key is only present in a dataset
        return 'dimensions' not in attributes


def normalize_slices(slices, shape):
    """ Normalize slices to shape.

    Normalize input, which can be a slice or a tuple of slices / ellipsis to
    be of same length as shape and be in bounds of shape.

    Args:
        slices (int or slice or ellipsis or tuple[int or slice or ellipsis]): slices to be normalized

    Returns:
        tuple[slice]: normalized slices (start and stop are both non-None)
        tuple[int]: which singleton dimensions should be squeezed out
    """
    type_msg = 'Advanced selection inappropriate. ' \
               'Only numbers, slices (`:`), and ellipsis (`...`) are valid indices (or tuples thereof)'

    if isinstance(slices, tuple):
        slices_lst = list(slices)
    elif isinstance(slices, (numbers.Number, slice, type(Ellipsis))):
        slices_lst = [slices]
    else:
        raise TypeError(type_msg)

    ndim = len(shape)
    if len([item for item in slices_lst if item != Ellipsis]) > ndim:
        raise TypeError("Argument sequence too long")
    elif len(slices_lst) < ndim and Ellipsis not in slices_lst:
        slices_lst.append(Ellipsis)

    normalized = []
    found_ellipsis = False
    squeeze = []
    for item in slices_lst:
        d = len(normalized)
        if isinstance(item, slice):
            normalized.append(slice_to_start_stop(item, shape[d]))
        elif isinstance(item, numbers.Number):
            squeeze.append(d)
            normalized.append(int_to_start_stop(int(item), shape[d]))
        elif isinstance(item, type(Ellipsis)):
            if found_ellipsis:
                raise ValueError("Only one ellipsis may be used")
            found_ellipsis = True
            while len(normalized) + (len(slices_lst) - d - 1) < ndim:
                normalized.append(slice(0, shape[len(normalized)]))
        else:
            raise TypeError(type_msg)
    return tuple(normalized), tuple(squeeze)



import unittest
import pickle
import sys
from shutil import rmtree

import numpy as np

try:
    import z5py
except ImportError:
    sys.path.append('..')
    import z5py


class TestPickle(unittest.TestCase):

    def setUp(self):
        self.shape = (100, 100, 100)
        self.ff = z5py.File('array.n5')

    def tearDown(self):
        try:
            rmtree('array.n5')
        except OSError:
            pass

    def test_pickle(self):
        dtypes = ('int8', 'int16', 'int32', 'int64',
                  'uint8', 'uint16', 'uint32', 'uint64',
                  'float32', 'float64')

        for dtype in dtypes:
            ds = self.ff.create_dataset('data_%s' % dtype,
                                        dtype=dtype,
                                        shape=self.shape,
                                        chunks=(10, 10, 10))
            in_array = 42 * np.ones(self.shape, dtype=dtype)
            ds[:] = in_array

            # test the original dataset
            out_array = ds[:]
            self.assertEqual(out_array.shape, in_array.shape)
            self.assertTrue(np.allclose(out_array, in_array))

            # pickle, unpickle and then test the datasets
            pickled = pickle.dumps(ds)
            ds_pickle = pickle.loads(pickled)
            out_pickle = ds_pickle[:]
            self.assertEqual(out_pickle.shape, in_array.shape)
            self.assertTrue(np.allclose(out_pickle, in_array))


if __name__ == '__main__':
    unittest.main()



from __future__ import print_function

import unittest
import json
import sys
from shutil import rmtree
from abc import ABC

import numpy as np

try:
    import z5py
except ImportError:
    sys.path.append('..')
    import z5py


# Just a dummy numpy encoder to test custom encoders
class NumpyEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        else:
            return super(NumpyEncoder, self).default(obj)


class DecoratingDecoder(json.JSONDecoder):
    def __init__(self, *args, **kwargs):
        json.JSONDecoder.__init__(self, object_hook=self.object_hook,
                                  *args, **kwargs)

    def object_hook(self, obj):
        return {k: 'a' + v + 'e' if isinstance(v, str) else v
                for k, v in obj.items()}


class Dummy:
    pass


class AttributesTestMixin(ABC):

    def setUp(self):
        self.shape = (100, 100, 100)

        self.root_file = z5py.File('array.%s' % self.data_format)
        self.root_file.create_dataset('ds', dtype='float32',
                                      shape=self.shape, chunks=(10, 10, 10))
        self.root_file.create_group('group')

    def tearDown(self):
        try:
            rmtree('array.%s' % self.data_format)
        except OSError:
            pass

    def check_attrs(self, attrs):
        self.assertFalse('not_an_attr' in attrs)

        attrs["a"] = 1
        attrs["b"] = [1, 2, 3]
        attrs["c"] = "whooosa"
        self.assertEqual(attrs["a"], 1)
        self.assertEqual(attrs["b"], [1, 2, 3])
        self.assertEqual(attrs["c"], "whooosa")

    def check_ds_attrs(self, attrs):
        for key in attrs.reserved_keys:
            with self.assertRaises(RuntimeError):
                attrs[key] = 5

            with self.assertRaises(RuntimeError):
                del attrs[key]

            self.assertIsNone(attrs.get(key))
            self.assertFalse(key in attrs)
            self.assertFalse(key in set(attrs))

    def test_attrs(self):

        # test file attributes
        f_attrs = self.root_file.attrs
        self.check_attrs(f_attrs)

        # test group attributes
        f_group = self.root_file["group"].attrs
        self.check_attrs(f_group)

        # test ds attributes
        f_ds = self.root_file["ds"].attrs
        self.check_attrs(f_ds)
        if not self.root_file.is_zarr:
            self.check_ds_attrs(f_ds)

    def test_custom_encoder(self):
        # check that we can't set arbitrary encoders
        with self.assertRaises(RuntimeError):
            z5py.set_json_encoder(Dummy)

        def _check():
            x = np.arange(10, dtype='int32')
            self.root_file.attrs['x'] = x
            y = self.root_file.attrs['x']
            self.assertTrue(np.allclose(x, y))

        # check that the encoder is used
        z5py.set_json_encoder(NumpyEncoder)
        _check()
        # check that the encoder is properly reset
        z5py.set_json_encoder(None)
        with self.assertRaises(TypeError):
            _check()

    # FIXME this fails in python 2 with, dunno why:
    # self.assertEqual(self.root_file.attrs['x'], 'abcde')
    # AssertionError: u'bcd' != 'abcde'
    @unittest.skipUnless(int(sys.version[0]) == 3, 'Json decoding seems to be different in python2')
    def test_custom_decoder(self):
        # check that we can't set arbitrary decoders
        with self.assertRaises(RuntimeError):
            z5py.set_json_decoder(Dummy)

        def _check():
            self.root_file.attrs['x'] = 'bcd'
            self.assertEqual(self.root_file.attrs['x'], 'abcde')

        # check that the decoder is used
        z5py.set_json_decoder(DecoratingDecoder)
        _check()
        # check that the decoder is properly reset
        z5py.set_json_decoder(None)
        with self.assertRaises(AssertionError):
            _check()


class TestAttributesZarr(AttributesTestMixin, unittest.TestCase):
    data_format = 'zarr'


class TestAttributesN5(AttributesTestMixin, unittest.TestCase):
    data_format = 'n5'


if __name__ == '__main__':
    unittest.main()



import unittest
import sys
import numpy as np
import os
from shutil import rmtree
from abc import ABC

try:
    import z5py
except ImportError:
    sys.path.append('..')
    import z5py


class DatasetTestMixin(ABC):
    def setUp(self):
        self.shape = (100, 100, 100)
        self.root_file = z5py.File('array.' + self.data_format,
                                   use_zarr_format=self.data_format == 'zarr')

        self.base_dtypes = [
            'int8', 'int16', 'int32', 'int64',
            'uint8', 'uint16', 'uint32', 'uint64',
            'float32', 'float64'
        ]
        self.dtypes = tuple(
            self.base_dtypes +
            [np.dtype(s) for s in self.base_dtypes] +
            [
                '<i1', '<i2', '<i4', '<i8',
                '<u1', '<u2', '<u4', '<u8',
                '<f4', '<f8'
            ] +
            [
                np.int8, np.int16, np.int32, np.int64,
                np.uint8, np.uint16, np.uint32, np.uint64,
                np.float32, np.float64
            ]
        )

    def tearDown(self):
        try:
            rmtree('array.' + self.data_format)
        except OSError:
            pass

    def check_array(self, result, expected, msg=None):
        self.assertEqual(result.shape, expected.shape, msg)
        self.assertTrue(np.allclose(result, expected), msg)

    def test_ds_open_empty(self):
        self.root_file.create_dataset('test',
                                      dtype='float32',
                                      shape=self.shape,
                                      chunks=(10, 10, 10))
        ds = self.root_file['test']
        out = ds[:]
        self.check_array(out, np.zeros(self.shape))

    def test_ds_dtypes(self):
        shape = (100, 100)
        chunks = (10, 10)
        for dtype in self.dtypes:
            ds = self.root_file.create_dataset('data_%s' % hash(dtype),
                                               dtype=dtype,
                                               shape=shape,
                                               chunks=chunks)
            in_array = np.random.rand(*shape).astype(dtype)
            ds[:] = in_array
            out_array = ds[:]
            self.check_array(out_array, in_array,
                             'datatype %s failed for format %s' % (self.data_format.title(),
                                                                   dtype))

    def check_ones(self, sliced_ones, expected_shape, msg=None):
        self.check_array(sliced_ones, np.ones(expected_shape, dtype=np.uint8), msg)

    def test_ds_simple_write(self):
        ds = self.root_file.create_dataset('ones', dtype=np.uint8,
                                           shape=self.shape, chunks=(10, 10, 10))
        ds[:] = np.ones(self.shape, np.uint8)

    def test_ds_indexing(self):
        ds = self.root_file.create_dataset('ones', dtype=np.uint8,
                                           shape=self.shape, chunks=(10, 10, 10))
        ds[:] = np.ones(self.shape, np.uint8)

        self.check_ones(ds[:], self.shape, 'full index failed')

        self.check_ones(ds[1, ...], (100, 100), 'trailing ellipsis failed')
        self.check_ones(ds[..., 1], (100, 100), 'leading ellipsis failed')
        self.check_ones(ds[1], (100, 100), 'implicit ellipsis failed')
        self.check_ones(ds[:, :, :, ...], self.shape, 'superfluous ellipsis failed')
        self.check_ones(ds[500:501, :, :], (0, 100, 100), 'out-of-bounds slice failed')
        self.check_ones(ds[-501:500, :, :], (0, 100, 100), 'negative out-of-bounds slice failed')

        self.check_ones(ds[1, :, :], (100, 100), 'integer index failed')
        self.check_ones(ds[-20:, :, :], (20, 100, 100), 'negative slice failed')

        self.assertEqual(ds[1, 1, 1], 1, 'point index failed')

        with self.assertRaises(ValueError):
            ds[500, :, :]
        with self.assertRaises(ValueError):
            ds[-500, :, :]
        with self.assertRaises(ValueError):
            ds[..., :, ...]
        with self.assertRaises(ValueError):
            ds[1, 1, slice(0, 100, 2)]
        with self.assertRaises(TypeError):
            ds[[1, 1, 1]]  # explicitly test behaviour different to h5py

        class NotAnIndex(object):
            pass

        with self.assertRaises(TypeError):
            ds[1, 1, NotAnIndex()]

    def test_ds_scalar_broadcast(self):
        for dtype in self.base_dtypes:
            ds = self.root_file.create_dataset('ones_%s' % dtype,
                                               dtype=dtype,
                                               shape=self.shape,
                                               chunks=(10, 10, 10))
            ds[:] = 1
            self.check_ones(ds[:], self.shape)

    def test_ds_scalar_broadcast_from_float(self):
        ds = self.root_file.create_dataset('ones', dtype=np.uint8,
                                           shape=self.shape, chunks=(10, 10, 10))
        ds[:] = float(1)
        self.check_ones(ds[:], self.shape),

    def test_ds_scalar_broadcast_from_bool(self):
        ds = self.root_file.create_dataset('ones', dtype=np.uint8,
                                           shape=self.shape, chunks=(10, 10, 10))
        ds[:] = True
        self.check_ones(ds[:], self.shape)

    def test_ds_set_with_arraylike(self):
        ds = self.root_file.create_dataset('ones', dtype=np.uint8,
                                           shape=self.shape, chunks=(10, 10, 10))
        ds[0, :2, :2] = [[1, 1], [1, 1]]
        self.check_ones(ds[0, :2, :2], (2, 2))

    def test_ds_set_from_float(self):
        ds = self.root_file.create_dataset('ones', dtype=np.uint8,
                                           shape=self.shape, chunks=(10, 10, 10))
        ds[:] = np.ones(self.shape, dtype=float)
        self.check_ones(ds[:], self.shape)

    def test_ds_set_from_bool(self):
        ds = self.root_file.create_dataset('ones', dtype=np.uint8,
                                           shape=self.shape, chunks=(10, 10, 10))
        ds[:] = np.ones(self.shape, dtype=bool)
        self.check_ones(ds[:], self.shape)

    def test_ds_fancy_broadcast_fails(self):
        ds = self.root_file.create_dataset('ones', dtype=np.uint8,
                                           shape=self.shape, chunks=(10, 10, 10))
        with self.assertRaises(ValueError):
            ds[0, :10, :10] = np.ones(10, dtype=np.uint8)

    def test_ds_write_object_fails(self):
        ds = self.root_file.create_dataset('ones', dtype=np.uint8,
                                           shape=self.shape, chunks=(10, 10, 10))

        class ArbitraryObject(object):
            pass

        with self.assertRaises(OSError):
            ds[0, 0, :2] = [ArbitraryObject(), ArbitraryObject()]

    def test_ds_write_flexible_fails(self):
        ds = self.root_file.create_dataset('ones', dtype=np.uint8,
                                           shape=self.shape, chunks=(10, 10, 10))
        with self.assertRaises(TypeError):
            ds[0, 0, 0] = "hey, you're not a number"

    def test_readwrite_multithreaded(self):
        for n_threads in (1, 2, 4, 8):
            ds = self.root_file.create_dataset('data_mthread_%i' % n_threads,
                                               dtype='float64',
                                               shape=self.shape,
                                               chunks=(10, 10, 10),
                                               n_threads=n_threads)
            in_array = np.random.rand(*self.shape)
            ds[:] = in_array
            out_array = ds[:]
            self.check_array(out_array, in_array)

    def test_create_nested_dataset(self):
        ds = self.root_file.create_dataset('group/sub_group/data',
                                           shape=self.shape,
                                           dtype='float64',
                                           chunks=(10, 10, 10))
        self.assertEqual(ds.path, os.path.join(self.root_file.path, 'group/sub_group/data'))

    def test_create_with_data(self):
        in_array = np.random.rand(*self.shape)
        ds = self.root_file.create_dataset('data', data=in_array)
        out_array = ds[:]
        self.check_array(out_array, in_array)

    def test_require_dataset(self):
        in_array = np.random.rand(*self.shape)
        self.root_file.require_dataset('data', data=in_array,
                                       shape=in_array.shape,
                                       dtype=in_array.dtype)
        ds = self.root_file.require_dataset('data',
                                            shape=in_array.shape,
                                            dtype=in_array.dtype)
        out_array = ds[:]
        self.check_array(out_array, in_array)

    def test_non_contiguous(self):
        ds = self.root_file.create_dataset('test',
                                           dtype='float32',
                                           shape=self.shape,
                                           chunks=(10, 10, 10))
        # make a non-contiguous 3d array of the correct shape (100)^3
        vol = np.arange(200**3).astype('float32').reshape((200, 200, 200))
        in_array = vol[::2, ::2, ::2]
        ds[:] = in_array
        out_array = ds[:]
        self.check_array(out_array, in_array, 'failed for non-contiguous data')

    def test_empty_chunk(self):
        ds = self.root_file.create_dataset('test',
                                           dtype='float32',
                                           shape=self.shape,
                                           chunks=(10, 10, 10))
        bb = np.s_[:10, :10, :10]
        if ds.is_zarr:
            chunk_path = os.path.join(ds.path, '0.0.0')
        else:
            chunk_path = os.path.join(ds.path, '0', '0', '0')
        ds[bb] = 0
        self.assertFalse(os.path.exists(chunk_path))
        ds[bb] = 1
        self.assertTrue(os.path.exists(chunk_path))
        ds[bb] = 0
        self.assertFalse(os.path.exists(chunk_path))

    def test_invalid_options(self):
        with self.assertRaises(RuntimeError):
            self.root_file.create_dataset('test1', shape=self.shape, dtype='float32',
                                          chunks=(10, 10, 10), compression='raw',
                                          level=5)
        with self.assertRaises(RuntimeError):
            self.root_file.create_dataset('test2', shape=self.shape, dtype='float32',
                                          chunks=(10, 10, 10), compression='bzip2',
                                          level=5, blub='blob')

    def test_readwrite_chunk(self):
        shape = (100, 100)
        chunks = (10, 10)
        for dtype in self.base_dtypes:
            ds = self.root_file.create_dataset('test_%s' % dtype, dtype=dtype,
                                               shape=shape, chunks=chunks,
                                               compression='raw')
            # test empty chunk
            out = ds.read_chunk((0, 0))
            self.assertEqual(out, None)

            # test read/write
            chunks_per_dim = ds.chunks_per_dimension
            for x in range(chunks_per_dim[0]):
                for y in range(chunks_per_dim[1]):
                    data = np.random.rand(*chunks)
                    if dtype not in ('float32', 'float64'):
                        data *= 128
                    data = data.astype(dtype)
                    ds.write_chunk((x, y), data)
                    out = ds.read_chunk((x, y))
                    self.assertEqual(data.shape, out.shape)
                    self.assertTrue(np.allclose(data, out))

    def test_read_direct(self):
        shape = (100, 100)
        chunks = (10, 10)

        ds = self.root_file.create_dataset('test', dtype='float64',
                                           shape=shape, chunks=chunks,
                                           compression='raw')

        # generate test data
        data = np.random.rand(*shape)
        ds[:] = data

        # test reading full dataset
        out = np.zeros(shape)
        ds.read_direct(out)
        self.assertTrue(np.allclose(out, data))

        # test reading with selection
        selection = np.s_[11:53, 67:84]
        out = np.zeros(shape)
        ds.read_direct(out, selection, selection)
        self.assertTrue(np.allclose(out[selection], data[selection]))

    def test_write_direct(self):
        shape = (100, 100)
        chunks = (10, 10)

        ds = self.root_file.create_dataset('test', dtype='float64',
                                           shape=shape, chunks=chunks,
                                           compression='raw')

        # generate test data
        data = np.random.rand(*shape)

        # test writing full dataset
        ds.write_direct(data)
        out = ds[:]
        self.assertTrue(np.allclose(out, data))

        # test writing with selection
        ds[:] = 0
        selection = np.s_[11:53, 67:84]
        ds.write_direct(data, selection, selection)
        out = ds[:]
        self.assertTrue(np.allclose(out[selection], data[selection]))

    def test_irregular_chunks(self):
        shape = (123, 54, 211)
        chunks = (13, 33, 22)

        ds = self.root_file.create_dataset('test', dtype='float64',
                                           shape=shape, chunks=chunks,
                                           compression='raw')
        data = np.random.rand(*shape)
        ds[:] = data
        out = ds[:]
        self.assertTrue(np.allclose(out, data))

    def test_nd(self):
        f = self.root_file
        for ndim in range(1, 6):
            size = 100 if ndim < 4 else 20
            shape = (size,) * ndim
            chunks = (10,) * ndim
            ds = f.create_dataset('test_%i' % ndim, dtype='float64',
                                  shape=shape, chunks=chunks, compression='raw')

            data = np.random.rand(*shape)
            ds[:] = data
            out = ds[:]
            self.assertTrue(np.allclose(out, data))

    def test_no_implicit_squeeze(self):
        arr = np.ones((5, 5, 5))
        ds = self.root_file.create_dataset('ds', data=arr)

        self.assertEqual(ds[:, 0:1, :].shape, arr[:, 0:1, :].shape)

    def test_no_implicit_squeeze_singleton(self):
        """Issue #102

        https://github.com/constantinpape/z5/issues/102
        """
        arr = np.ones((5, 5, 5))
        ds = self.root_file.create_dataset('ds', data=arr)
        self.assertEqual(
            ds[0:1, 0:1, 0:1].shape,
            arr[0:1, 0:1, 0:1].shape,
        )

    def test_explicit_squeeze(self):
        """Issue #103

        https://github.com/constantinpape/z5/issues/103
        """
        arr = np.full((5, 4, 3), 1)
        ds = self.root_file.create_dataset('ds543', data=arr)
        self.assertEqual(ds[:, 1, :].shape, arr[:, 1, :].shape)

        self.assertNotIsInstance(ds[1, 1, 1], np.ndarray)

    def test_singleton_dtype(self):
        """Issue #102

        https://github.com/constantinpape/z5/issues/102
        """
        arr = np.ones((5, 5, 5))
        ds = self.root_file.create_dataset('ds', data=arr)
        self.assertEqual(type(ds[1, 1, 1]), type(arr[1, 1, 1]))

    def test_broadcast_empty(self):
        """Issue #107

        https://github.com/constantinpape/z5/issues/107
        """
        ds = self.root_file.create_dataset('test', shape=(100, 100), chunks=(25, 25),
                                           dtype='uint8', compression='raw')
        ds[:20, :20] = 1
        out = ds[:]
        self.assertTrue(np.allclose(out[:20, :20], 1))

    def test_empty_chunks_non_aligned_write(self):
        """Issue #106

        https://github.com/constantinpape/z5/issues/106
        """
        ds = self.root_file.create_dataset(name='test', shape=(128,), chunks=(32,),
                                           compression='raw', dtype='uint8')

        inp = np.ones((100,), dtype='uint8')
        inp[90:100] = 0
        ds[:100] = inp
        # last chunk should be empty, but this is not the case if buffer was not
        # cleared correctly
        out = ds[-32:]
        self.assertTrue(np.allclose(out, 0))


class TestZarrDataset(DatasetTestMixin, unittest.TestCase):
    data_format = 'zarr'

    def test_varlen(self):
        shape = (100, 100)
        chunks = (10, 10)
        ds = self.root_file.create_dataset('varlen', dtype='float64',
                                           shape=shape, chunks=chunks,
                                           compression='raw')
        with self.assertRaises(RuntimeError):
            ds.write_chunk((0, 0), np.random.rand(10), True)


class TestN5Dataset(DatasetTestMixin, unittest.TestCase):
    data_format = 'n5'

    @unittest.skipIf(sys.version_info.major < 3, "This fails in python 2")
    def test_ds_array_to_format(self):
        for dtype in self.base_dtypes:
            ds = self.root_file.create_dataset('data_%s' % hash(dtype),
                                               dtype=dtype,
                                               shape=self.shape,
                                               chunks=(10, 10, 10))
            in_array = 42 * np.ones((10, 10, 10), dtype=dtype)
            ds[:10, :10, :10] = in_array

            path = os.path.join(os.path.dirname(ds.attrs.path), '0', '0', '0')
            with open(path, 'rb') as f:
                read_from_file = np.array([byte for byte in f.read()], dtype='int8')

            converted_data = ds.array_to_format(in_array)

            self.assertEqual(len(read_from_file), len(converted_data))
            self.assertTrue(np.allclose(read_from_file, converted_data))

    def test_varlen(self):
        shape = (100, 100)
        chunks = (10, 10)
        ds = self.root_file.create_dataset('varlen', dtype='float64',
                                           shape=shape, chunks=chunks,
                                           compression='raw')

        # max_len = 100
        max_len = 10
        chunks_per_dim = ds.chunks_per_dimension
        for x in range(chunks_per_dim[0]):
            for y in range(chunks_per_dim[1]):
                test_data = np.random.rand(np.random.randint(0, max_len))
                ds.write_chunk((x, y), test_data, True)
                out = ds.read_chunk((x, y))
                self.assertEqual(test_data.shape, out.shape)
                self.assertTrue(np.allclose(test_data, out))


if __name__ == '__main__':
    unittest.main()



import unittest
import os
from shutil import rmtree
from abc import ABC

import sys
try:
    import z5py
except ImportError:
    sys.path.append('..')
    import z5py


class FileTestMixin(ABC):
    def setUp(self):
        self.path = 'file.%s' % self.data_format

    def tearDown(self):
        try:
            rmtree(self.path)
        except OSError:
            pass

    def test_context_manager(self):
        self.assertFalse(os.path.exists(self.path))

        with z5py.File(self.path) as f:
            self.assertIsInstance(f, z5py.File)
            if self.data_format == 'n5':
                self.assertFalse(f.is_zarr)
            else:
                self.assertTrue(f.is_zarr)
        self.assertTrue(os.path.exists(self.path))

    def test_extension_detect(self):
        self.assertFalse(os.path.exists(self.path))

        f = z5py.File(self.path, use_zarr_format=None)
        if self.data_format == 'n5':
            self.assertFalse(f.is_zarr)
        else:
            self.assertTrue(f.is_zarr)

    def test_wrong_ext_fails(self):
        is_n5 = self.data_format == 'n5'
        with self.assertRaises(RuntimeError):
            z5py.File(self.path, use_zarr_format=is_n5)


class TestZarrFile(FileTestMixin, unittest.TestCase):
    data_format = 'zarr'

    def test_direct_constructor(self):
        self.assertFalse(os.path.exists(self.path))
        f = z5py.ZarrFile(self.path)
        self.assertTrue(f.is_zarr)


class TestZrFile(FileTestMixin, unittest.TestCase):
    data_format = 'zr'

    def test_direct_constructor(self):
        self.assertFalse(os.path.exists(self.path))
        f = z5py.ZarrFile(self.path)
        self.assertTrue(f.is_zarr)


class TestN5File(FileTestMixin, unittest.TestCase):
    data_format = 'n5'

    def test_direct_constructor(self):
        self.assertFalse(os.path.exists(self.path))
        f = z5py.N5File(self.path)
        self.assertFalse(f.is_zarr)


if __name__ == '__main__':
    unittest.main()



import os
import numbers
import json

import numpy as np

from . import _z5py
from .attribute_manager import AttributeManager, get_json_encoder
from .shape_utils import normalize_slices, rectify_shape
from .shape_utils import get_default_chunks, is_group


class Dataset:
    """ Dataset for access to data on disc.

    Should not be instantiated directly, but rather
    be created or opened via ``create_dataset``, ``require_dataset`` or
    the ``[]`` operator of File or Group.
    """

    _dtype_dict = {np.dtype('uint8'): 'uint8',
                   np.dtype('uint16'): 'uint16',
                   np.dtype('uint32'): 'uint32',
                   np.dtype('uint64'): 'uint64',
                   np.dtype('int8'): 'int8',
                   np.dtype('int16'): 'int16',
                   np.dtype('int32'): 'int32',
                   np.dtype('int64'): 'int64',
                   np.dtype('float32'): 'float32',
                   np.dtype('float64'): 'float64'}

    _zarr_dtype_dict = {np.dtype('uint8'): '|u1',
                        np.dtype('uint16'): '<u2',
                        np.dtype('uint32'): '<u4',
                        np.dtype('uint64'): '<u8',
                        np.dtype('int8'): '|i1',
                        np.dtype('int16'): '<i2',
                        np.dtype('int32'): '<i4',
                        np.dtype('int64'): '<i8',
                        np.dtype('float32'): '<f4',
                        np.dtype('float64'): '<f8'}

    # TODO for now we hardcode all compressors
    # but we should instead check which ones are present
    # (similar to nifty WITH_CPLEX, etc.)

    #: Compression libraries supported by zarr format
    compressors_zarr = ['raw', 'blosc', 'zlib', 'bzip2', 'gzip']
    #: Default compression for zarr format
    zarr_default_compressor = 'blosc'

    # TODO support blosc in n5
    #: Compression libraries supported by n5 format
    compressors_n5 = ['raw', 'gzip', 'bzip2', 'xz', 'lz4']
    #: Default compression for n5 format
    n5_default_compressor = 'gzip'

    def __init__(self, path, dset_impl, n_threads=1):
        self._impl = dset_impl
        self._attrs = AttributeManager(path, self._impl.is_zarr)
        self.path = path
        self.n_threads = n_threads

    @staticmethod
    def _to_zarr_compression_options(compression, compression_options):
        if compression == 'blosc':
            default_opts = {'id': 'blosc', 'cname': 'lz4', 'clevel': 5, 'shuffle': 1}
        elif compression == 'zlib':
            default_opts = {'id': 'zlib', 'level': 5}
        elif compression == 'gzip':
            default_opts = {'id': 'gzip', 'level': 5}
        elif compression == 'bzip2':
            default_opts = {'id': 'bz2', 'level': 5}
        elif compression == 'raw':
            default_opts = {}
        else:
            raise RuntimeError("Compression %s is not supported in zarr format" % compression)

        # check for invalid options
        extra_args = set(compression_options) - set(default_opts)
        if extra_args:
            raise RuntimeError("Invalid options for %s compression: %s" % (compression, ' '.join(list(extra_args))))

        # return none for raw compression
        if not default_opts:
            return None

        # update the default options
        default_opts.update(compression_options)
        return default_opts

    def _read_zarr_compression_options(self):
        opts = {}
        with open(os.path.join(self.path, '.zarray'), 'r') as f:
            zarr_opts = json.load(f)['compressor']
        if zarr_opts is None:
            opts['compression'] = 'raw'
        elif zarr_opts['id'] == 'blosc':
            opts['compression'] = 'blosc'
            opts['level'] = zarr_opts['clevel']
            opts['shuffle'] = zarr_opts['shuffle']
            opts['codec'] = zarr_opts['cname']
        elif zarr_opts['id'] == 'zlib':
            opts['compression'] = 'zlib'
            opts['level'] = zarr_opts['level']
        elif zarr_opts['id'] == 'bz2':
            opts['compression'] = 'bzip2'
            opts['level'] = zarr_opts['level']
        elif zarr_opts['id'] == 'gzip':
            opts['compression'] = 'gzip'
            opts['level'] = zarr_opts['level']
        return opts

    @staticmethod
    def _to_n5_compression_options(compression, compression_options):
        if compression == 'gzip':
            default_opts = {'type': 'gzip', 'level': 5}
            level_key = None  # for several compressors, we need to change the parameter key
        elif compression == 'bzip2':
            default_opts = {'type': 'bzip2', 'level': 5}
            level_key = 'blockSize'
        elif compression == 'raw':
            default_opts = {'type': 'raw'}
            level_key = None
        elif compression == 'xz':
            default_opts = {'type': 'xz', 'level': 6}
            level_key = 'preset'
        elif compression == 'lz4':
            default_opts = {'type': 'lz4', 'level': 6}
            level_key = 'blockSize'
        # TODO blosc in n5
        # elif compression == 'blosc':
        #    default_opts = {'type': 'blosc', 'codec': 'lz4', 'level': 5, 'shuffle': 1}
        #    level_key = None
        else:
            raise RuntimeError("Compression %s is not supported in n5 format" % compression)

        # check for invalid options
        extra_args = set(compression_options) - set(default_opts)
        if extra_args:
            raise RuntimeError("Invalid options for %s compression: %s" % (compression, ' '.join(list(extra_args))))

        # update the default options
        default_opts.update(compression_options)

        # rectify the level option for compressions which have a key different than level
        if level_key is not None:
            default_opts[level_key] = default_opts.pop('level')

        return default_opts

    def _read_n5_compression_options(self):
        opts = {}
        with open(os.path.join(self.path, 'attributes.json'), 'r') as f:
            n5_opts = json.load(f)
        # old compression scheme
        if 'compressionType' in n5_opts:
            ctype = n5_opts['compressionType']
            new_compression = False
        # new compression scheme
        else:
            ctype = n5_opts['compression']['type']
            new_compression = True

        if ctype == 'raw':
            opts['compression'] = 'raw'
        elif ctype == 'gzip':
            opts['compression'] = 'gzip'
            opts['level'] = n5_opts['compression']['level'] if new_compression else 5
        elif ctype == 'bzip2':
            opts['compression'] = 'bzip2'
            opts['level'] = n5_opts['compression']['blockSize'] if new_compression else 5
        elif ctype == 'xz':
            opts['compression'] = 'xz'
            opts['level'] = n5_opts['compression']['preset'] if new_compression else 6
        elif ctype == 'lz4':
            opts['compression'] = 'lz4'
            opts['level'] = n5_opts['compression']['blockSize'] if new_compression else 6
        # TODO blosc in n5
        # elif n5_opts['id'] == 'blosc':
        #     opts['compression'] = 'blosc'
        #     opts['level'] = n5_opts['clevel']
        #     opts['shuffle'] = n5_opts['shuffle']
        #     opts['codec'] = n5_opts['cname']
        return opts

    @staticmethod
    def _create_dataset_zarr(path, dtype, shape, chunks,
                             compression, compression_options,
                             fill_value):
        os.makedirs(path)
        params = {'dtype': Dataset._zarr_dtype_dict[np.dtype(dtype)],
                  'shape': shape,
                  'chunks': chunks,
                  'fill_value': fill_value,
                  'zarr_format': 2,
                  'order': 'C',
                  'filters': None,
                  'compressor': Dataset._to_zarr_compression_options(compression,
                                                                     compression_options)}
        with open(os.path.join(path, '.zarray'), 'w') as f:
            json.dump(params, f, cls=get_json_encoder())

    @staticmethod
    def _create_dataset_n5(path, dtype, shape, chunks,
                           compression, compression_options):
        os.makedirs(path)
        params = {'dataType': Dataset._dtype_dict[np.dtype(dtype)],
                  'dimensions': shape[::-1],
                  'blockSize': chunks[::-1],
                  'compression': Dataset._to_n5_compression_options(compression,
                                                                    compression_options)}
        with open(os.path.join(path, 'attributes.json'), 'w') as f:
            json.dump(params, f, cls=get_json_encoder())

    # NOTE in contrast to h5py, we also check that the chunks match
    # this is crucial, because different chunks can lead to subsequent incorrect
    # code when relying on chunk-aligned access for parallel writing
    @classmethod
    def _require_dataset(cls, path,
                         shape, dtype,
                         chunks, n_threads,
                         is_zarr, mode, **kwargs):
        if os.path.exists(path):

            if is_group(path, is_zarr):
                raise TypeError("Incompatible object (Group) already exists")

            ds = cls(path, _z5py.open_dataset(path, mode), n_threads)
            if shape != ds.shape:
                raise TypeError("Shapes do not match (existing (%s) vs new (%s))" % (', '.join(map(str, ds.shape)),
                                                                                     ', '.join(map(str, shape))))
            if chunks is not None:
                if chunks != ds.chunks:
                    raise TypeError("Chunks do not match (existing (%s) vs new (%s))" % (', '.join(map(str, ds.chunks)),
                                                                                         ', '.join(map(str, chunks))))
            if dtype is not None:
                if np.dtype(dtype) != ds.dtype:
                    raise TypeError("Datatypes do not match (existing %s vs new %s)" % str(ds.dtype), str(dtype))
            return ds

        else:
            # pop all kwargs that are not compression options
            data = kwargs.pop('data', None)
            compression = kwargs.pop('compression', None)
            fillvalue = kwargs.pop('fillvalue', 0)
            return cls._create_dataset(path, shape, dtype, data=data,
                                       chunks=chunks, compression=compression,
                                       fillvalue=fillvalue, n_threads=n_threads,
                                       compression_options=kwargs,
                                       is_zarr=is_zarr, mode=mode)

    @classmethod
    def _create_dataset(cls, path, shape, dtype,
                        data=None, chunks=None,
                        compression=None,
                        fillvalue=0, n_threads=1,
                        compression_options={},
                        is_zarr=True,
                        mode=None):
        # check if this dataset already exists
        if os.path.exists(path):
            raise RuntimeError("Cannot create dataset (name already exists)")

        # check shape, dtype and data
        have_data = data is not None
        if have_data:
            if shape is None:
                shape = data.shape
            elif shape != data.shape:
                raise ValueError("Shape Tuple is incompatible with data")
            if dtype is None:
                dtype = data.dtype
            # NOTE In contrast to h5py, we don't do automatic type conversion
            elif np.dtype(dtype) != data.dtype:
                raise ValueError("Datatype is incompatible with data")
        else:
            if shape is None:
                raise TypeError("One of shape or data must be specified")
            # NOTE In contrast to h5py (and numpy), we DO NOT have float64
            # as default data type, but require a data type if no data is given
            if dtype is None:
                raise TypeError("One of dtype or data must be specified")
        parsed_dtype = np.dtype(dtype)

        # get default chunks if necessary
        # NOTE in contrast to h5py, datasets are chunked
        # by default, with chunk size ~ 64**3
        if chunks is None:
            chunks = get_default_chunks(shape)
        # check chunks have the same len than the shape
        if len(chunks) != len(shape):
            raise RuntimeError("Chunks %s must have same length as shape %s" % (str(chunks),
                                                                                str(shape)))
        # limit chunks to shape
        chunks = tuple(min(ch, sh) for ch, sh in zip(chunks, shape))

        # check compression / get default compression
        # if no compression is given
        if compression is None:
            compression = cls.zarr_default_compressor if is_zarr else cls.n5_default_compressor
        else:
            valid_compression = compression in cls.compressors_zarr if is_zarr else\
                compression in cls.compressors_n5
            if not valid_compression:
                raise ValueError("Compression filter \"%s\" is unavailable" % compression)

        # get and check compression
        if is_zarr and compression not in cls.compressors_zarr:
            compression = cls.zarr_default_compressor
        elif not is_zarr and compression not in cls.compressors_n5:
            compression = cls.n5_default_compressor

        if is_zarr:
            if parsed_dtype not in cls._zarr_dtype_dict:
                raise ValueError("Invalid data type {} for zarr dataset".format(dtype))
            cls._create_dataset_zarr(path, parsed_dtype, shape, chunks,
                                     compression, compression_options, fillvalue)
        else:
            if parsed_dtype not in cls._dtype_dict:
                raise ValueError("Invalid data type {} for N5 dataset".format(repr(dtype)))
            cls._create_dataset_n5(path, parsed_dtype, shape, chunks,
                                   compression, compression_options)

        # get the dataset and write data if necessary
        ds = cls(path, _z5py.open_dataset(path, mode), n_threads)
        if have_data:
            ds[:] = data
        return ds

    @classmethod
    def _open_dataset(cls, path, mode):
        return cls(path, _z5py.open_dataset(path, mode))

    @property
    def is_zarr(self):
        """ Flag to indicate zarr or n5 format of this dataset.
        """
        return self._impl.is_zarr

    @property
    def attrs(self):
        """ The ``AttributeManager`` of this dataset.
        """
        return self._attrs

    @property
    def shape(self):
        """ Shape of this dataset.
        """
        return tuple(self._impl.shape)

    @property
    def ndim(self):
        """ Number of dimensions of this dataset.
        """
        return self._impl.ndim

    @property
    def size(self):
        """ Size (total number of elements) of this dataset.
        """
        return self._impl.size

    @property
    def chunks(self):
        """ Chunks of this dataset.
        """
        return tuple(self._impl.chunks)

    @property
    def dtype(self):
        """ Datatype of this dataset.
        """
        return np.dtype(self._impl.dtype)

    @property
    def chunks_per_dimension(self):
        """ Number of chunks in each dimension of this dataset.
        """
        return self._impl.chunks_per_dimension

    @property
    def number_of_chunks(self):
        """ Total number of chunks of this dataset.
        """
        return self._impl.number_of_chunks

    @property
    def compression(self):
        opts = self.compression_opts
        compression = opts['compression']
        return compression if compression != 'raw' else None

    @property
    def compression_opts(self):
        """ Compression library options of this dataset.
        """
        return self._read_zarr_compression_options() if self._impl.is_zarr else \
            self._read_n5_compression_options()

    def __len__(self):
        return self._impl.len

    def index_to_roi(self, index):
        """ Convert index to region of interest.

        Convert an index, which can be a slice or a tuple of slices / ellipsis to a
        region of interest. The roi consists of the region offset and the region shape.

        Args:
            index (slice or tuple): index into dataset.

        Returns:
            tuple: offset of the region of interest.
            tuple: shape of the region of interest.
            tuple: which dimensions should be squeezed out
        """
        normalized, to_squeeze = normalize_slices(index, self.shape)
        return (
            tuple(norm.start for norm in normalized),
            tuple(
                0 if norm.start is None else norm.stop - norm.start for norm in normalized
            ),
            to_squeeze
        )

    # most checks are done in c++
    def __getitem__(self, index):
        roi_begin, shape, to_squeeze = self.index_to_roi(index)
        out = np.empty(shape, dtype=self.dtype)
        if 0 not in shape:
            _z5py.read_subarray(self._impl,
                                out, roi_begin,
                                n_threads=self.n_threads)

        # todo: this probably has more copies than necessary
        if len(to_squeeze) == len(shape):
            return out.flatten()[0]
        elif to_squeeze:
            return out.squeeze(to_squeeze)
        else:
            return out

    # most checks are done in c++
    def __setitem__(self, index, item):
        roi_begin, shape, _ = self.index_to_roi(index)
        if 0 in shape:
            return

        # broadcast scalar
        if isinstance(item, (numbers.Number, np.number)):
            _z5py.write_scalar(self._impl, roi_begin,
                               list(shape), item,
                               str(self.dtype), self.n_threads)
            return

        try:
            item_arr = np.asarray(item, self.dtype, order='C')
        except ValueError as e:
            if any(s in str(e) for s in ('invalid literal for ', 'could not convert')):
                bad_dtype = np.asarray(item).dtype
                raise TypeError("No conversion path for dtype: " + repr(bad_dtype))
            else:
                raise
        except TypeError as e:
            if 'argument must be' in str(e):
                raise OSError("Can't prepare for writing data (no appropriate function for conversion path)")
            else:
                raise

        item_arr = rectify_shape(item_arr, shape)
        _z5py.write_subarray(self._impl,
                             item_arr,
                             roi_begin,
                             n_threads=self.n_threads)

    def find_minimum_coordinates(self, dim):
        """ Find coordinates of chunk with smallest coordinate along dimension.

        Only considers chunks that contain data.

        Args:
            dim (int): query dimension.
        Returns:
            tuple: start coordinates of the chunk.
        """
        return tuple(self._impl.findMinimumCoordinates(dim))

    def find_maximum_coordinates(self, dim):
        """ Find coordinates of chunk with largest coordinate along dimension.

        Only considers chunks that contain data.

        Args:
            dim (int): query dimension.
        Returns:
            tuple: start coordinates of the chunk.
        """
        return self._impl.findMaximumCoordinates(dim)

    def read_direct(self, dest, source_sel=None, dest_sel=None):
        """ Wrapper to improve similarity to h5py. Reads from the dataset to ``dest``, using ``read_subarray``.

        Args:
            dest (array) destination object into which the read data is written to.
            dest_sel (slice array) selection of data to write to ``dest``. Defaults to the whole range of ``dest``.
            source_sel (slice array) selection in dataset to read from. Defaults to the whole range of the dataset.
        Spaces, defined by ``source_sel`` and ``dest_sel`` must be in the same size but dont need to have the same
        offset
        """
        if source_sel is None:
            source_sel = tuple(slice(0, sh) for sh in self.shape)
        if dest_sel is None:
            dest_sel = tuple(slice(0, sh) for sh in dest.shape)
        start = [s.start for s in source_sel]
        stop = [s.stop for s in source_sel]
        dest[dest_sel] = self.read_subarray(start, stop)

    def write_direct(self, source, source_sel=None, dest_sel=None):
        """ Wrapper to improve similarity to h5py. Writes to the dataset from ``source``, using ``write_subarray``.

        Args:
            source (array) source object from which the written data is obtained.
            source_sel (slice array) selection of data to write from ``source``. Defaults to the whole range of
            ``source``.
            dest_sel (slice array) selection in dataset to write to. Defaults to the whole range of the dataset.
        Spaces, defined by ``source_sel`` and ``dest_sel`` must be in the same size but dont need to have the same
        offset
        """
        if dest_sel is None:
            dest_sel = tuple(slice(0, sh) for sh in self.shape)
        if source_sel is None:
            source_sel = tuple(slice(0, sh) for sh in source.shape)
        start = [s.start for s in dest_sel]
        self.write_subarray(start, source[source_sel])

    # expose the impl write subarray functionality
    def write_subarray(self, start, data):
        """ Write subarray to dataset.

        ``data`` is written to region of interest, defined by ``start``
        and the shape of ``data``. The region of interest must be in
        bounds of the dataset and the datatype must agree with the dataset.

        Args:
            start (tuple): offset of the roi to write.
            data (np.ndarray): data to write; shape determines the roi shape.
        """
        _z5py.write_subarray(self._impl,
                             np.require(data, requirements='C'),
                             list(start),
                             n_threads=self.n_threads)

    # expose the impl read subarray functionality
    def read_subarray(self, start, stop):
        """ Read subarray from region of interest.

        Region of interest is defined by ``start`` and ``stop``
        and must be in bounds of the dataset.

        Args:
            start (tuple): start coordinates of the roi.
            stop (tuple): stop coordinates of the roi.

        Returns:
            np.ndarray
        """
        shape = tuple(sto - sta for sta, sto in zip(start, stop))
        out = np.empty(shape, dtype=self.dtype)
        _z5py.read_subarray(self._impl, out, start, n_threads=self.n_threads)
        return out

    def array_to_format(self, array):
        """ Convert array to serialization.

        Convert an array to the (1d) binary data that would be serialized to disc
        for the format of the dataset.

        Args:
            array (np.ndarray): array to be converted to serialization.

        Returns:
            np.ndarray
        """
        if array.ndim != self.ndim:
            raise RuntimeError("Array needs to be of same dimension as dataset")
        if np.dtype(array.dtype) != np.dtype(self.dtype):
            raise RuntimeError("Array needs to have same dtype as dataset")
        return _z5py.convert_array_to_format(self._impl,
                                             np.require(array, requirements='C'))

    def chunk_exists(self, chunk_indices):
        """ Check if chunk has data.

        Check for the given indices if the chunk has data.

        Args:
            chunk_indices (tuple): chunk indices.

        Returns:
            bool
        """
        return self._impl.chunkExists(chunk_indices)

    def write_chunk(self, chunk_indices, data, varlen=False):
        """ Write single chunk

        Args:
            chunk_indices (tuple): indices of the chunk to write to
            data (np.ndarray): data to be written
            varlen (bool): write this chunk in varlen mode; only supported in n5
                (default: False)
        """
        if self.is_zarr and varlen:
            raise RuntimeError("Varlength chunks are not supported in zarr")
        if not varlen:
            pass
            # TODO make sure that shapes agree
            # chunk_shape = ''
            # if chunk_shape != data.shape:
            #   raise ValueError("")
        _z5py.write_chunk(self._impl, chunk_indices, data, varlen)

    def read_chunk(self, chunk_indices):
        """ Read single chunk

        Args:
            chunk_indices (tuple): indices of the chunk to write to
        Returns
            np.ndarray
        """
        chunk_reader = getattr(_z5py, 'read_chunk_%s' % self._impl.dtype)
        # FIXME this is super hacky, but for some reason returning none in pybind raises
        # an exception for integer types:
        # "TypeError: int() argument must be a string,
        # a bytes-like object or a number, not 'NoneType"
        # so for now we catch the exception
        # for float types, it doesn't, but returns an array with `nan`
        try:
            out = chunk_reader(self._impl, chunk_indices)
        except TypeError:
            return None
        return None if np.isnan(out).any() else out



import unittest
import os
import numpy as np
from shutil import rmtree
from abc import ABC

import sys
try:
    import z5py
except ImportError:
    sys.path.append('..')
    import z5py


class GroupTestMixin(ABC):

    def setUp(self):
        self.shape = (100, 100, 100)

        self.root_file = z5py.File('array.%s' % self.data_format)
        g = self.root_file.create_group('test')
        g.create_dataset('test',
                         dtype='float32',
                         shape=self.shape,
                         chunks=(10, 10, 10))

    def tearDown(self):
        try:
            rmtree('array.%s' % self.data_format)
        except OSError:
            pass

    def test_contains(self):
        self.assertTrue('test' in self.root_file)
        self.assertTrue('/test' in self.root_file)
        self.assertTrue('/test/test' in self.root_file)
        self.assertFalse('fake' in self.root_file)

    def test_open_group(self):
        g = self.root_file['test']
        ds = g['test']
        out = ds[:]
        self.assertEqual(out.shape, self.shape)
        self.assertTrue((out == 0).all())

    def test_open_dataset(self):
        ds = self.root_file['test/test']
        out = ds[:]
        self.assertEqual(out.shape, self.shape)
        self.assertTrue((out == 0).all())

    def test_group(self):
        g = self.root_file.create_group('group')
        ds = g.create_dataset('data',
                              dtype='float32',
                              shape=self.shape,
                              chunks=(10, 10, 10))
        in_array = np.random.rand(*self.shape).astype('float32')
        ds[:] = in_array
        out_array = ds[:]
        self.assertEqual(out_array.shape, in_array.shape)
        self.assertTrue(np.allclose(out_array, in_array))

    def test_create_nested_group(self):
        self.root_file.create_group('foo/bar/baz')
        g = self.root_file['foo/bar/baz']
        self.assertEqual(g.path, os.path.join(self.root_file.path, 'foo/bar/baz'))

    def test_require_group(self):
        self.root_file.require_group('group')
        g = self.root_file.require_group('group')
        self.assertEqual(g.path, os.path.join(self.root_file.path, 'group'))

    def test_delete(self):
        self.assertTrue('test' in self.root_file)
        self.assertTrue('test/test' in self.root_file)
        del self.root_file['test']
        self.assertFalse('test' in self.root_file)
        self.assertFalse('test/test' in self.root_file)

    def test_leading_slash(self):
        # make sure that we can read slashes
        g = self.root_file['/test']
        ds1 = g['/test']
        ds2 = self.root_file['/test/test']
        self.assertTrue(np.allclose(ds1[:], ds2[:]))

    def test_require_dataset_with_compression_options(self):
        """Issue #98

        https://github.com/constantinpape/z5/issues/98
        """
        g = self.root_file['/test']
        ds = g.require_dataset('new', shape=(100, 100), chunks=(10, 10),
                               dtype='uint8', compression='gzip',
                               level=4)
        self.assertTrue(os.path.exists(ds.path))
        compression_opts = ds.compression_opts
        self.assertEqual(compression_opts['level'], 4)

    # cf. https://github.com/constantinpape/z5/issues/100
    @unittest.skipIf(sys.platform.startswith('win'), 'path encoding not compatible with windows')
    def test_visititems(self):
        """ Issue #121

        https://github.com/constantinpape/z5/issues/121
        """
        f = self.root_file
        f.create_group('g1')
        f.create_dataset('g1/d1', shape=(10, 10), dtype='uint8')
        f.create_dataset('g1/d2', shape=(12, 12), dtype='uint8')

        names = []

        def visitor(name, obj):
            names.append(name)

        f.visititems(visitor)
        names = set(names)
        expected_names = {'g1', 'g1/d1', 'g1/d2', 'test', 'test/test'}
        self.assertEqual(names, expected_names)

        names = []
        g = f['g1']
        g.visititems(visitor)
        names = set(names)
        expected_names = {'d1', 'd2'}
        self.assertEqual(names, expected_names)


class TestGroupZarr(GroupTestMixin, unittest.TestCase):
    data_format = 'zr'


class TestGroupN5(GroupTestMixin, unittest.TestCase):
    data_format = 'n5'


if __name__ == '__main__':
    unittest.main()



import os
from shutil import rmtree

try:
    from collections.abc import Mapping
except ImportError:
    from collections import Mapping

from ._z5py import create_group, FileMode
from .dataset import Dataset
from .attribute_manager import AttributeManager
from .shape_utils import is_group


class Group(Mapping):
    """ Group inside of a z5py container.

    Corresponds to a directory on the filesystem.
    Supports python dict api.
    Should not be instantiated directly, but rather be created
    or opened via the `create_group`, `request_group` or `[]` operators
    of Group or File.
    """

    # the python / h5py file modes and the corresponding internal types
    # as far as I can tell there is no difference between 'w-' and 'x',
    # so for now they get mapped to the same internal type

    #: available modes for opening files. these correspond to the ``h5py`` file modes
    file_modes = {'a': FileMode.a, 'r': FileMode.r,
                  'r+': FileMode.r_p, 'w': FileMode.w,
                  'w-': FileMode.w_m, 'x': FileMode.w_m}

    def __init__(self, path, is_zarr=True, mode='a'):
        if mode not in self.file_modes:
            raise ValueError("Invalid file mode: %s" % mode)

        self.mode = mode
        self._internal_mode = self.file_modes[mode]
        self._permissions = FileMode(self._internal_mode)
        self.path = path
        self.is_zarr = is_zarr
        self._attrs = AttributeManager(path, is_zarr)

    #
    # Magic Methods, Attributes, Keys, Contains
    #

    def __iter__(self):
        for name in os.listdir(self.path):
            if os.path.isdir(os.path.join(self.path, name)):
                yield name

    def __len__(self):
        counter = 0
        for _ in self:
            counter += 1
        return counter

    def __contains__(self, name):
        return super(Group, self).__contains__(name.lstrip('/'))

    # TODO can use `util.removeDataset` here
    def __delitem__(self, name):
        if not self._permissions.can_write():
            raise ValueError("Cannot call delete on file not opened with write permissions.")
        path_ = os.path.join(self.path, name)
        if not os.path.exists(path_):
            raise KeyError("%s does not exist" % name)
        rmtree(path_)

    def __getitem__(self, name):
        """ Access group or dataset in the container.

        Fails if no dataset or group of the specified name exists.

        Args:
            name (str): name of group or dataset in container.

        Returns:
            ``Group`` or ``Dataset``.
        """
        path = os.path.join(self.path, name.lstrip('/'))
        if not os.path.isdir(path):
            raise KeyError("Key %s does not exist" % name)

        if is_group(path, self.is_zarr):
            return Group._open_group(path, self.is_zarr, self.mode)
        else:
            return Dataset._open_dataset(path, self._internal_mode)

    @property
    def attrs(self):
        """ Access additional attributes.

        Returns:
            ``AttributeManager``.
        """
        return self._attrs

    #
    # Group functionality
    #

    @classmethod
    def _create_group(cls, path, is_zarr, mode):
        create_group(path, is_zarr, cls.file_modes[mode])
        return cls(path, is_zarr, mode)

    @classmethod
    def _open_group(cls, path, is_zarr, mode):
        return cls(path, is_zarr, mode)

    def create_group(self, name):
        """ Create a new group.

        Create new (sub-)group of the group.
        Fails if a group of this name already exists.

        Args:
            name (str): name of the new group.

        Returns:
            ``Group``: group of the requested name.
        """
        if not self._permissions.can_write():
            raise ValueError("Cannot create group with read-only permissions.")
        if name in self:
            raise KeyError("Group %s already exists" % name)
        path = os.path.join(self.path, name.lstrip('/'))
        return Group._create_group(path, self.is_zarr, self.mode)

    def require_group(self, name):
        """ Require group.

        Require that a group of the given name exists.
        The group will be created if it does not already exist.

        Args:
            name (str): name of the required group.

        Returns:
            ``Group``: group of the requested name.
        """
        path = os.path.join(self.path, name.lstrip('/'))
        if os.path.exists(path):
            if not is_group(path, self.is_zarr):
                raise TypeError("Incompatible object (Dataset) already exists")
            return Group._open_group(path, self.is_zarr, self.mode)
        else:
            return self.create_group(name)

    #
    # Dataset functionality
    #

    def create_dataset(self, name,
                       shape=None, dtype=None,
                       data=None, chunks=None,
                       compression=None, fillvalue=0,
                       n_threads=1, **compression_options):
        """ Create a new dataset.

        Create a new dataset in the group. Syntax and behaviour similar to the
        corresponding ``h5py`` functionality.
        In contrast to ``h5py``, there is no option to store a dataset without chunking
        (if no chunks are given default values, suitable for the dimension of the dataset, will be used).
        Also, if a dataset is created with data and a dtype that is different
        from the data's is specified, the function throws a RuntimeError, instead
        of converting the data.

        Args:
            name (str): name of the new dataset.
            shape (tuple): shape of the new dataset. If no shape is given,
                the ``data`` argument must be given. (default: None).
            dtype (str or np.dtpye): datatype of the new dataset. If no dtype is given,
                the ``data`` argument must be given (default: None).
            data (np.ndarray): data used to infer shape, dtype and fill the dataset
                upon creation (default: None).
            chunks (tuple): chunk sizes of the new dataset. If no chunks are given,
                a suitable default value for the number of dimensions will be used (default: None).
            compression (str): name of the compression library used to compress chunks.
                If no compression is given, the default for the current format is used (default: None).
            fillvalue (float): fillvalue for empty chunks (only zarr) (default: 0).
            n_threads (int): number of threads used for chunk I/O (default: 1).
            **compression_options: options for the compression library.

        Returns:
            ``Dataset``: the new dataset.
        """

        if not self._permissions.can_write():
            raise ValueError("Cannot create dataset with read-only permissions.")
        if name in self:
            raise KeyError("Dataset %s is already existing." % name)
        path = os.path.join(self.path, name.lstrip('/'))
        return Dataset._create_dataset(path, shape, dtype,
                                       data, chunks, compression,
                                       fillvalue, n_threads,
                                       compression_options,
                                       self.is_zarr, self._internal_mode)

    def require_dataset(self, name, shape,
                        dtype=None, chunks=None,
                        n_threads=1, **kwargs):
        """ Require dataset.

        Require dataset in the group.
        Will create the dataset if it does not exist, otherwise returns
        existing dataset. If the dataset already exists, consistency with the
        arguments ``shape``, ``dtype`` (if given) and ``chunks`` (if given) is enforced.

        Args:
            name (str): name of the dataset.
            shape (tuple): shape of the dataset.
            dtype (str or np.dtpye): datatype of dataset (default: None).
            chunks (tuple): chunk sizes of the dataset (default: None).
            n_threads (int): number of threads used for chunk I/O (default: 1).
            **kwargs: additional arguments that will only be used for creation
                if the dataset does not exist.

        Returns:
            ``Dataset``: the required dataset.
        """
        if not self._permissions.can_write():
            raise ValueError("Cannot create dataset with read-only permissions.")
        path = os.path.join(self.path, name.lstrip('/'))
        return Dataset._require_dataset(path, shape, dtype, chunks,
                                        n_threads, self.is_zarr, self._internal_mode,
                                        **kwargs)

    def visititems(self, func, root=None):
        """ Recursively visit names and objects in this group.

        You supply a callable (function, method or callable object); it
        will be called exactly once for each link in this group and every
        group below it. Your callable must conform to the signature:

            func(<member name>, <object>) => <None or return value>

        Returning None continues iteration, returning anything else stops
        and immediately returns that value from the visit method.  No
        particular order of iteration within groups is guranteed.

        calls the function @param func with the found items, appended to @param path

        Example:

        # Get a list of all datasets in the file
        >>> mylist = []
        >>> def func(name, obj):
        ...     if isinstance(obj, Dataset):
        ...         mylist.append(name)
        ...
        >>> f = File('foo.n5')
        >>> f.visititems(func)
        """
        root = self.path if root is None else root
        for name, obj in self.items():
            # the name needs to be relative to the root object visititems was called on
            name = os.path.relpath(os.path.join(self.path, name), root)
            func_ret = func(name, obj)
            if func_ret is not None:
                return func_ret
            if isinstance(obj, Group):
                obj.visititems(func, root=root)



import unittest
import os
from shutil import rmtree
import numpy as np

import sys
try:
    import z5py
except ImportError:
    sys.path.append('..')
    import z5py


# We just test for the n5 ending, as this is independent of
# the data format
class TestPermissions(unittest.TestCase):
    def setUp(self):
        self.path1 = 'array1.n5'
        self.path2 = 'array2.n5'
        f = z5py.File(self.path1)
        f.create_dataset('data', dtype='uint8', shape=(10, 10), chunks=(10, 10))

    def tearDown(self):
        try:
            rmtree(self.path1)
        except OSError:
            pass
        try:
            rmtree(self.path2)
        except OSError:
            pass

    # test read permissions ('r')
    def test_read_permission(self):
        self.assertTrue(os.path.exists(self.path1))
        self.assertFalse(os.path.exists(self.path2))

        # test that a new file can't be created
        with self.assertRaises(OSError):
            z5py.File(self.path2, mode='r')

        # open existing file
        f = z5py.File(self.path1, mode='r')
        self.assertFalse(f.is_zarr)

        # TODO would be nice to have consistent exceptions here
        # test that a new group can't be created
        with self.assertRaises(ValueError):
            f.create_group('g')

        # test that a new dataset can't be created
        # with self.assertRaises(RuntimeError):
        with self.assertRaises(ValueError):
            f.create_dataset('ds', dtype='uint8', shape=(10, 10), chunks=(10, 10))

        # test that we cannot write to an existing dataset
        ds = f['data']
        with self.assertRaises(ValueError):
            ds[:] = np.zeros((10, 10), dtype='uint8')

    # generic test for all default operations
    def generic_test(self, f):
        self.assertFalse(f.is_zarr)
        f.create_group('g')
        ds = f.create_dataset('ds', dtype='uint8', shape=(10, 10), chunks=(10, 10))
        ds[:] = np.zeros((10, 10), dtype='uint8')

    # test write permissions ('w')
    def test_write_permission(self):
        self.assertTrue(os.path.exists(self.path1))
        self.assertFalse(os.path.exists(self.path2))

        # we should be able to do all the default operations in write
        f = z5py.File(self.path2, mode='w')
        self.generic_test(f)

        # make sure that append truncates
        f = z5py.File(self.path1, mode='w')
        in_dir = os.listdir(self.path1)
        self.assertEqual(len(in_dir), 0)

    # test append permissions ('a')
    def test_append_permission(self):
        self.assertTrue(os.path.exists(self.path1))
        self.assertFalse(os.path.exists(self.path2))

        # we should be able to do all the default operations in append
        f = z5py.File(self.path2, mode='a')
        self.generic_test(f)

    # test read-plus permissions ('r+')
    def test_readplus_permission(self):
        self.assertTrue(os.path.exists(self.path1))
        self.assertFalse(os.path.exists(self.path2))

        # we should be able to do all the default operations in read-plus (except creation)
        f = z5py.File(self.path1, mode='r+')
        self.generic_test(f)

        # we are not allowed to do creation in readplus
        with self.assertRaises(OSError):
            z5py.File(self.path2, mode='r+')

    # test write-minus permissions ('w-')
    def test_writeminus_permission(self):
        self.assertTrue(os.path.exists(self.path1))
        self.assertFalse(os.path.exists(self.path2))

        # we should be able to do all the default operations in read-minus (except opening file that already exists)
        f = z5py.File(self.path2, mode='w-')
        self.generic_test(f)

        # we are not allowed to do creation in readplus
        with self.assertRaises(OSError):
            z5py.File(self.path1, mode='w-')


if __name__ == '__main__':
    unittest.main()



import sys
import unittest
import numpy as np
import os
from concurrent import futures
from shutil import rmtree

try:
    import z5py
except ImportError:
    sys.path.append('..')
    import z5py


class TestUtil(unittest.TestCase):
    tmp_dir = './tmp_dir'
    shape = (100, 100, 100)
    chunks = (10, 10, 10)

    def setUp(self):
        if not os.path.exists(self.tmp_dir):
            os.mkdir(self.tmp_dir)

    def tearDown(self):
        try:
            rmtree(self.tmp_dir)
        except OSError:
            pass

    def test_copy_dataset_with_roi(self):
        from z5py.util import copy_dataset
        in_path = os.path.join(self.tmp_dir, 'in.n5')
        out_path = os.path.join(self.tmp_dir, 'out.n5')

        in_file = z5py.File(in_path, use_zarr_format=False)
        out_file = z5py.File(out_path, use_zarr_format=False)
        # create input dataset
        ds_in = in_file.create_dataset('data', dtype='float32',
                                       shape=self.shape, chunks=self.chunks,
                                       compression='gzip')
        # write test data
        data = np.arange(ds_in.size).reshape(ds_in.shape).astype(ds_in.dtype)
        ds_in[:] = data

        # define roi
        roi = np.s_[5:45, 9:83, 10:60]
        out_of_roi_mask = np.ones(self.shape, dtype='bool')
        out_of_roi_mask[roi] = False
        roi_shape = tuple(rr.stop - rr.start for rr in roi)

        # copy dataset with roi
        copy_dataset(in_path, out_path, 'data', 'data_roi', n_threads=8,
                     roi=roi, fit_to_roi=False)
        ds_out = out_file['data_roi']
        data_out = ds_out[:]

        self.assertEqual(data_out.shape, data.shape)
        self.assertTrue(np.allclose(data_out[roi], data[roi]))
        self.assertTrue(np.allclose(data_out[out_of_roi_mask], 0))

        # copy dataset with roi and fit_to_roi
        copy_dataset(in_path, out_path, 'data', 'data_roi_fit', n_threads=8,
                     roi=roi, fit_to_roi=True)
        ds_out = out_file['data_roi_fit']
        data_out = ds_out[:]

        self.assertEqual(data_out.shape, roi_shape)
        self.assertTrue(np.allclose(data_out, data[roi]))

    @unittest.skipUnless(futures, "Needs 3rd party concurrent.futures in python 2")
    def test_copy_dataset_default(self):
        from z5py.util import copy_dataset
        in_path = os.path.join(self.tmp_dir, 'in.n5')
        out_path = os.path.join(self.tmp_dir, 'out.n5')

        # create input file
        in_file = z5py.File(in_path, use_zarr_format=False)
        ds_in = in_file.create_dataset('data', dtype='float32',
                                       shape=self.shape, chunks=self.chunks,
                                       compression='gzip')
        # write test data
        data = np.arange(ds_in.size).reshape(ds_in.shape).astype(ds_in.dtype)
        ds_in[:] = data

        # copy_dataset for different out blocks
        out_file = z5py.File(out_path, use_zarr_format=False)
        new_chunks = (20, 20, 20)

        # NOTE we can only choose out blocks that align with the chunks
        # because otherwise we run into issues due to not thread safe blocking
        for out_blocks in (None, (40, 40, 40), (60, 60, 60)):
            ds_str = 'none' if out_blocks is None else '_'.join(map(str, out_blocks))
            ds_name = 'data_%s' % ds_str
            copy_dataset(in_path, out_path, 'data', ds_name,
                         chunks=new_chunks,
                         block_shape=out_blocks,
                         n_threads=8)
            # make sure that new data agrees
            ds_out = out_file[ds_name]
            data_out = ds_out[:]
            self.assertEqual(data_out.shape, data.shape)
            self.assertEqual(ds_out.chunks, new_chunks)
            self.assertTrue(np.allclose(data, data_out))

    @unittest.skipUnless(futures, "Needs 3rd party concurrent.futures in python 2")
    def test_copy_dataset_custom(self):
        from z5py.util import copy_dataset
        in_path = os.path.join(self.tmp_dir, 'in.n5')
        out_path = os.path.join(self.tmp_dir, 'out.n5')

        # create input file
        in_file = z5py.File(in_path, use_zarr_format=False)
        ds_in = in_file.create_dataset('data', dtype='float32',
                                       shape=self.shape, chunks=self.chunks,
                                       compression='gzip')
        # write test data
        data = np.arange(ds_in.size).reshape(ds_in.shape).astype(ds_in.dtype)
        ds_in[:] = data
        # copy_dataset
        new_chunks = (20, 20, 20)
        for compression in ('raw', 'gzip'):
            for dtype in ('float64', 'int32', 'uint32'):
                ds_name = 'ds_%s_%s' % (compression, dtype)
                copy_dataset(in_path, out_path,
                             'data', ds_name,
                             chunks=new_chunks,
                             n_threads=8,
                             compression=compression,
                             dtype=dtype)
                # make sure that new data agrees
                out_file = z5py.File(out_path, use_zarr_format=False)
                ds_out = out_file[ds_name]
                data_out = ds_out[:]
                self.assertEqual(data_out.shape, data.shape)
                self.assertEqual(ds_out.chunks, new_chunks)
                self.assertTrue(np.allclose(data, data_out))

    # TODO finish blocking tests
    def simple_blocking(self, shape, block_shape):
        blocking = []
        ndim = len(shape)
        for x in range(0, shape[0], block_shape[0]):
            blocking.append((x, min(x + block_shape[0], shape[0])))
            if ndim > 1:
                block = blocking.pop()
                for y in range(0, shape[1], block_shape[1]):
                    blocking.append(block + min(y + block_shape[1], shape[1]))

    def _test_blocking(self):
        from z5py.util import blocking
        n_reps = 10

        for dim in range(1, 6):
            for _ in range(n_reps):
                shape = tuple(np.random.randint(0, 1000) for ii in range(dim))
                block_shape = tuple(min(np.random.randint(0, 100), sh)
                                    for ii, sh in zip(range(dim), shape))
                blocking1 = [(block.start, block.stop)
                             for block in blocking(shape, block_shape)]
                blocking2 = self.simple_blocking(shape, block_shape)
                sorted(blocking1)
                sorted(blocking2)
                self.assertEqual(blocking1, blocking2)

    def test_remove_trivial_chunks(self):
        from z5py.util import remove_trivial_chunks
        path = './tmp_dir/data.n5'
        f = z5py.File(path)
        shape = (100, 100)
        chunks = (10, 10)

        a = np.zeros(shape)
        a[10:20, 10:20] = 1
        a[20:30, 20:30] = 2
        a[50:60, 50:60] = np.arange(100).reshape(chunks)

        ds = f.create_dataset('data', dtype='float64',
                              shape=shape, chunks=chunks)

        ds[:] = a
        remove_trivial_chunks(ds, n_threads=4)
        b = ds[:]

        self.assertTrue(np.allclose(b[10:20, 10:20], 0))
        self.assertTrue(np.allclose(b[20:30, 20:30], 0))
        self.assertTrue(np.allclose(b[50:60, 50:60],
                                    np.arange(100).reshape(chunks)))

        ds[:] = a
        remove_trivial_chunks(ds, n_threads=4, remove_specific_value=1)
        c = ds[:]

        self.assertTrue(np.allclose(c[10:20, 10:20], 0))
        self.assertTrue(np.allclose(c[20:30, 20:30], 2))
        self.assertTrue(np.allclose(c[50:60, 50:60],
                                    np.arange(100).reshape(chunks)))

    def test_unique(self):
        from z5py.util import unique
        path = './tmp_dir/data.n5'
        f = z5py.File(path)
        shape = (100, 100)
        chunks = (10, 10)

        ds = f.create_dataset('data', dtype='int32',
                              shape=shape, chunks=chunks)
        data = np.random.randint(0, 100, size=shape).astype('int32')
        ds[:] = data

        exp_uniques, exp_counts = np.unique(data, return_counts=True)
        uniques = unique(ds, n_threads=4)
        self.assertTrue(np.allclose(uniques, exp_uniques))

        uniques, counts = unique(ds, n_threads=4, return_counts=True)
        self.assertTrue(np.allclose(uniques, exp_uniques))
        self.assertTrue(np.allclose(counts, exp_counts))

    def test_remove_dataset(self):
        from z5py.util import remove_dataset
        path = './tmp_dir/data.n5'
        f = z5py.File(path)
        shape = (100, 100)
        chunks = (10, 10)

        ds = f.create_dataset('data', dtype='float64',
                              data=np.ones(shape), chunks=chunks)
        remove_dataset(ds, 4)
        self.assertFalse(os.path.exists(os.path.join(path, 'data')))


if __name__ == '__main__':
    unittest.main()
